from itertools import chain
from pathlib import Path
from typing import Any, Callable, Iterable, List, Optional, TypeVar, Union

import pyarrow.dataset
import pyarrow.fs
import pyarrow.parquet
from pyarrow.dataset import ParquetFileFragment

from edsnlp import registry
from edsnlp.core.lazy_collection import LazyCollection
from edsnlp.data.base import BaseReader, BaseWriter
from edsnlp.data.converters import (
    FILENAME,
    get_dict2doc_converter,
    get_doc2dict_converter,
)
from edsnlp.utils.collections import dl_to_ld, flatten, flatten_once, ld_to_dl
from edsnlp.utils.file_system import FileSystem, normalize_fs_path


class ParquetReader(BaseReader):
    DATA_FIELDS = ("dataset",)

    def __init__(
        self,
        path: Union[str, Path],
        *,
        read_in_worker: bool,
        filesystem: Optional[FileSystem] = None,
    ):
        super().__init__()
        # Either the filesystem has not been passed
        # or the path is a URL (e.g. s3://) => we need to infer the filesystem
        self.fs, self.path = normalize_fs_path(filesystem, path)
        self.read_in_worker = read_in_worker
        self.dataset: pyarrow.dataset.FileSystemDataset = (  # type: ignore
            pyarrow.dataset.dataset(
                self.path,
                format="parquet",
                filesystem=self.fs,
            )
        )

    def read_main(self):
        fragments: List[ParquetFileFragment] = self.dataset.get_fragments()
        if self.read_in_worker:
            # read in worker -> each task is a file to read from
            return ((f, f.metadata.num_rows) for f in fragments)
        else:
            # read in worker -> each task is a non yet parsed line
            return (
                (line, 1)
                for f in fragments
                # for line in make_longer(dl_to_ld(f.to_table().to_pydict()))
                for line in dl_to_ld(f.to_table().to_pydict())
            )

    def read_worker(self, tasks):
        if self.read_in_worker:
            tasks = list(
                chain.from_iterable(
                    dl_to_ld(task.to_table().to_pydict()) for task in tasks
                )
            )
        return tasks


T = TypeVar("T")


class ParquetWriter(BaseWriter):
    def __init__(
        self,
        path: Union[str, Path],
        num_rows_per_file: int,
        overwrite: bool,
        write_in_worker: bool,
        accumulate: bool = True,
        filesystem: Optional[FileSystem] = None,
    ):
        super().__init__()
        self.fs, self.path = normalize_fs_path(filesystem, path)

        # Check that filesystem has the same protocol as indicated by path
        self.fs.makedir(self.path, create_parents=True)
        dataset: pyarrow.dataset.FileSystemDataset = (  # type: ignore
            pyarrow.dataset.dataset(
                self.path,
                format="parquet",
                filesystem=self.fs,
            )
        )
        self.fs.makedirs(self.path, exist_ok=True)
        if len(list(dataset.files)):
            if not overwrite:
                raise FileExistsError(
                    f"Directory {self.path} already exists and is not empty. "
                    "Use overwrite=True to overwrite."
                )
            for file in dataset.files:
                self.fs.rm_file(file)
        self.fs = filesystem
        self.write_in_worker = write_in_worker
        self.batch = []
        self.num_rows_per_file = num_rows_per_file
        self.closed = False
        self.finalized = False
        self.accumulate = accumulate
        if not self.accumulate:
            self.finalize = super().finalize

    def write_worker(self, records, last=False):
        # Results will contain a batches of samples ready to be written (or None if
        # write_in_worker is True) and they have already been written.
        results = []
        count = 0

        for rec in records:
            if isinstance(rec, dict):
                rec.pop(FILENAME, None)

        # While there is something to write
        greedy = last or not self.accumulate
        while len(records) or greedy and len(self.batch):
            n_to_fill = self.num_rows_per_file - len(self.batch)
            self.batch.extend(records[:n_to_fill])
            records = records[n_to_fill:]
            if greedy or len(self.batch) >= self.num_rows_per_file:
                fragment = pyarrow.Table.from_pydict(ld_to_dl(flatten(self.batch)))  # type: ignore
                count += len(self.batch)
                self.batch = []
                if self.write_in_worker:
                    pyarrow.parquet.write_to_dataset(
                        table=fragment,
                        root_path=self.path,
                        filesystem=self.fs,
                    )
                    fragment = None
                results.append(fragment)
        return results, count

    def finalize(self):
        if not self.finalized:
            self.finalized = True
            return self.write_worker([], last=True)

    def write_main(self, fragments: Iterable[List[Union[pyarrow.Table, Path]]]):
        for table in flatten_once(fragments):
            if not self.write_in_worker:
                pyarrow.parquet.write_to_dataset(
                    table=table,
                    root_path=self.path,
                    filesystem=self.fs,
                )
        return pyarrow.dataset.dataset(self.path, format="parquet", filesystem=self.fs)


@registry.readers.register("parquet")
def read_parquet(
    path: Union[str, Path],
    converter: Optional[Union[str, Callable]] = None,
    *,
    read_in_worker: bool = False,
    filesystem: Optional[FileSystem] = None,
    **kwargs,
) -> LazyCollection:
    """
    The ParquetReader (or `edsnlp.data.read_parquet`) reads a directory of parquet files
    (or a single file) and yields documents.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)
    doc_iterator = edsnlp.data.read_parquet("path/to/parquet", converter="omop")
    annotated_docs = nlp.pipe(doc_iterator)
    ```

    !!! note "Generator vs list"

        `edsnlp.data.read_parquet` returns a
        [LazyCollection][edsnlp.core.lazy_collection.LazyCollection].
        To iterate over the documents multiple times efficiently or to access them by
        index, you must convert it to a list

        ```{ .python .no-check }
        docs = list(edsnlp.data.read_parquet("path/to/parquet", converter="omop"))
        ```

    Parameters
    ----------
    path: Union[str, Path]
        Path to the directory containing the parquet files (will recursively look for
        files in subdirectories). Supports any filesystem supported by pyarrow.
    read_in_worker: bool
        Whether to read the files in the worker or in the main process.
    filesystem: Optional[AbstractFileSystem] = None,
        The filesystem to use to write the files. If None, the filesystem will be
        inferred from the path (e.g. `s3://` will use S3).
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the parquet rows of the data source to Doc objects
        These are documented on the [Converters](/data/converters) page.
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.

    Returns
    -------
    LazyCollection
    """
    data = LazyCollection(
        reader=ParquetReader(
            path,
            read_in_worker=read_in_worker,
            filesystem=filesystem,
        )
    )
    if converter:
        converter, kwargs = get_dict2doc_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)
    return data


@registry.writers.register("parquet")
def write_parquet(
    data: Union[Any, LazyCollection],
    path: Union[str, Path],
    *,
    write_in_worker: bool = False,
    num_rows_per_file: int = 1024,
    overwrite: bool = False,
    filesystem: Optional[FileSystem] = None,
    accumulate: bool = True,
    converter: Optional[Union[str, Callable]] = None,
    **kwargs,
) -> None:
    """
    `edsnlp.data.write_parquet` writes a list of documents as a parquet dataset.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)

    doc = nlp("My document with entities")

    edsnlp.data.write_parquet([doc], "path/to/parquet")
    ```

    !!! warning "Overwriting files"

        By default, `write_parquet` will raise an error if the directory already exists
        and contains parquet files. This is to avoid overwriting existing annotations.
        To allow overwriting existing files, use `overwrite=True`.

    Parameters
    ----------
    data: Union[Any, LazyCollection],
        The data to write (either a list of documents or a LazyCollection).
    path: Union[str, Path]
        Path to the directory containing the parquet files (will recursively look for
        files in subdirectories). Supports any filesystem supported by pyarrow.
    num_rows_per_file: int
        The maximum number of documents to write in each parquet file.
    overwrite: bool
        Whether to overwrite existing directories.
    write_in_worker: bool
        Whether to write the files in the workers or in the main process.
    accumulate: bool
        Whether to accumulate the results sent to the writer by workers until the
        batch is full or the writer is finalized. If False, each file will not be larger
        than the size of the batches it receives. This option requires that the writer
        is finalized before the end of the processing, which may not be compatible with
        some backends, such as `spark`.

        If `write_in_worker` is True, documents will be accumulated in each worker but
        not across workers, therefore leading to a larger number of files.
    filesystem: Optional[AbstractFileSystem] = None,
        The filesystem to use to write the files. If None, the filesystem will be
        inferred from the path (e.g. `s3://` will use S3).
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the documents to dictionary objects before writing
        them as Parquet rows. These are documented on the [Converters](/data/converters)
        page.
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.
    """

    data = LazyCollection.ensure_lazy(data)
    if converter:
        converter, kwargs = get_doc2dict_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)

    return data.write(
        ParquetWriter(
            path,
            num_rows_per_file=num_rows_per_file,
            overwrite=overwrite,
            write_in_worker=write_in_worker,
            accumulate=accumulate,
            filesystem=filesystem,
        )
    )



from edsnlp.core import registry

from .adicap import AdicapMatcher
from .patterns import adicap_prefix, base_code

DEFAULT_CONFIG = dict(
    pattern=base_code,
    prefix=adicap_prefix,
    window=500,
    attr="TEXT",
    label="adicap",
    span_setter={"ents": True, "adicap": True},
)

create_component = registry.factory.register(
    "eds.adicap",
    assigns=["doc.ents", "doc.spans"],
)(AdicapMatcher)



from ..terms import ASYMPTOMATIC, BRAIN, HEART, PERIPHERAL

acronym = dict(
    source="acronym",
    regex=[
        r"\bAOMI\b",
        r"\bACOM\b",
        r"\bTAO\b",
        r"\bSAPL\b",
        r"\bOACR\b",
        r"\bOVCR\b",
        r"\bSCS\b",
        r"\bTVP\b",
        r"\bCAPS\b",
        r"\bMTEV\b",
        r"\bPTT\b",
        r"\bMAT\b",
        r"\bSHU\b",
    ],
    regex_attr="TEXT",
)

other = dict(
    source="other",
    regex=[
        r"\bbuerger",
        r"takayasu",
        r"\bhorton",
        r"wegener",
        r"churg.{1,10}strauss",
        r"\bsneddon",
        r"budd.chiari",
        r"infarctus.{1,5}(renal|splenique|polaire|pulmo)",
        r"ulcere.{1,5}arter",
        r"syndrome.?hemolytique.{1,8}uremique",
        r"granulomatose.{1,10}polyangeite",
        r"occlusion.{1,10}(artere|veine).{1,20}retine",
        r"syndrome.{1,20}anti.?phospho",
        r"embolie.{1,5}pulm",
    ],
    regex_attr="NORM",
)

with_localization = dict(
    source="with_localization",
    regex=[
        r"angiopathie",
        r"arteriopathies.{1,5}obliterante",
        r"gangren",
        r"claudication",
        r"dissection.{1,10}(aort|arter)",
        r"tromboangeit",
        r"tromboarterit",
        r"(pontage|angioplastie).{1,10}(\bfem|\bpop|\bren|\bjamb)",
        r"arterite",
        r"(ischemie|infarctus).{1,10}mesenterique",
        r"endarteriectomie",
        r"vascularite",
        r"occlusion.{1,10}terminaisons? carotid",
        r"cryoglobulinemie",
        r"colite.{1,5}ischemi",
        r"embole.{1,10}cholesterol",
        r"purpura.?thrombopenique.?idiopa",
        r"micro.?angiopathie.?thrombotique",
    ],
    exclude=[
        dict(
            regex=BRAIN + HEART + ASYMPTOMATIC + [r"inr\srecommande\ssous\savk"],
            window=(-8, 8),
            limit_to_sentence=False,
        ),
    ],
    regex_attr="NORM",
)

thrombosis = dict(
    source="thrombosis",
    regex=[
        r"thrombos",
        r"thrombol[^y]",
        r"thrombophi",
        r"thrombi[^n]",
        r"thrombus",
        r"thrombectomi",
        r"thrombo.?embo",
        r"phlebit",
    ],
    exclude=[
        dict(
            regex=BRAIN + HEART + ["superficiel", "\biv\b", "intra.?vein"],
            window=(-15, 15),
            limit_to_sentence=False,
        ),
        dict(
            regex=[
                "pre",
                "anti",
                "bilan",
            ],
            window=-4,
        ),
    ],
    regex_attr="NORM",
)


ischemia = dict(
    source="ischemia",
    regex=[
        r"ischemi",
    ],
    exclude=[
        dict(
            regex=BRAIN + HEART,
            window=(-7, 7),
        ),
    ],
    assign=[
        dict(
            name="peripheral",
            regex="(" + r"|".join(PERIPHERAL) + ")",
            window=15,
        ),
    ],
    regex_attr="NORM",
)

ep = dict(
    source="ep",
    regex=r"\bEP(?![\w\./-])",
    regex_attr="TEXT",
    exclude=[
        dict(
            regex=[
                r"fibreux",
                r"retin",
                r"\bfove",
                r"\boct\b",
                r"\bmacula",
                r"prosta",
                r"\bip\b",
                r"protocole",
                r"seance",
                r"echange",
                r"ritux",
                r"ivig",
                r"ig.?iv",
                r"\bctc",
                r"corticoide",
                r"serum",
                r"\bcure",
                r"plasma",
                r"mensuel",
                r"semaine",
                r"serologi",
                r"espaces.porte",
                r"projet",
                r"bolus",
            ],
            window=(-25, 25),
            limit_to_sentence=False,
            regex_attr="NORM",
        ),
        dict(
            regex=[r"rdv", r"les", r"des", r"angine"],
            window=(-3, 0),
            regex_attr="NORM",
        ),
    ],
)

hypertension = dict(
    source="main",
    regex=[
        r"\bhta\b",
        r"hyper.?tension.?arte",
        r"hyper.?tendu",
        r"hyper.?tension.?essenti",
        r"hypertensi",
    ],
    exclude=dict(
        regex="(pulmo|porta)",
        window=3,
    ),
)

default_patterns = [
    acronym,
    other,
    with_localization,
    thrombosis,
    ep,
    ischemia,
    hypertension,
]



results_ckd = dict(
    has_match=[
        True,
        False,
        True,
        True,
        False,
        True,
        False,
        True,
        True,
        True,
        False,
    ],
    detailled_status="PRESENT",
    assign=8 * [None] + [{"stage": "IV"}, {"dfg": 30}, None],
    texts=[
        "Patient atteint d'une glomérulopathie.",
        "Patient atteint d'une tubulopathie aigüe.",
        "Patient transplanté rénal",
        "Présence d'une insuffisance rénale aigüe sur chronique",
        "Le patient a été dialysé",  # ponctuelle
        "Le patient est dialysé chaque lundi",  # chronique
        "Présence d'une IRC",  # severity non mentionned
        "Présence d'une IRC sévère",
        "Présence d'une IRC de classe IV",
        "Présence d'une IRC avec DFG à 30",  # severe
        "Présence d'une maladie rénale avec DFG à 110",  # no renal failure
    ],
)



main_pattern = dict(
    source="main",
    regex=[
        r"hemiplegi",
        r"tetraplegi",
        r"quadriplegi",
        r"paraplegi",
        r"neuropathie.{1,25}motrice.{1,30}type [5V]",
        r"charcot.?marie.?tooth",
        r"locked.?in",
        r"syndrome.{1,5}(enfermement|verrouillage)|(desafferen)",
        r"paralysie.{1,10}hemicorps",
        r"paralysie.{1,10}jambe",
        r"paralysie.{1,10}membre",
        r"paralysie.{1,10}cote",
        r"paralysie.{1,5}cerebrale.{1,5}spastique",
    ],
    regex_attr="NORM",
)

acronym = dict(
    source="acronym",
    regex=[
        r"\bLIS\b",
        r"\bNMSH\b",
    ],
    regex_attr="TEXT",
)

default_patterns = [
    main_pattern,
    acronym,
]



from typing import List, Optional, Tuple

from spacy.tokens import Doc

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import BaseComponent

from . import patterns


class AccentsConverter(BaseComponent):
    """
    Normalises accents, using a same-length strategy.

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline object.
    name : Optional[str]
        The component name.
    accents : List[Tuple[str, str]]
        List of accentuated characters and their transcription.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol] = None,
        name: Optional[str] = "spaces",
        *,
        accents: List[Tuple[str, str]] = patterns.accents,
    ):
        super().__init__(nlp, name)
        self.translation_table = str.maketrans(
            "".join(accent_group for accent_group, _ in accents),
            "".join(rep * len(accent_group) for accent_group, rep in accents),
        )

    def __call__(self, doc: Doc) -> Doc:
        """
        Remove accents from spacy `NORM` attribute.

        Parameters
        ----------
        doc : Doc
            The spaCy `Doc` object.

        Returns
        -------
        Doc
            The document, with accents removed in `Token.norm_`.
        """

        for token in doc:
            token.norm_ = token.norm_.translate(self.translation_table)

        return doc



from typing import Any, Dict, Union

from typing_extensions import Literal

from edsnlp.core import PipelineProtocol, registry
from edsnlp.pipes.core.terminology.terminology import TerminologyMatcher

from ...base import SpanSetterArg
from .patterns import get_patterns

DEFAULT_CONFIG = dict(
    attr="NORM",
    ignore_excluded=False,
    ignore_space_tokens=False,
    term_matcher="exact",
    term_matcher_config={},
    pattern_config=dict(languages=["FRE"], sources=None),
    label="umls",
    span_setter={"ents": True, "umls": True},
)


@registry.factory.register(
    "eds.umls",
    assigns=["doc.ents", "doc.spans"],
)
def create_component(
    nlp: PipelineProtocol,
    name: str = "umls",
    *,
    attr: Union[str, Dict[str, str]] = "NORM",
    ignore_excluded: bool = False,
    ignore_space_tokens: bool = False,
    term_matcher: Literal["exact", "simstring"] = "exact",
    term_matcher_config: Dict[str, Any] = {},
    pattern_config: Dict[str, Any] = dict(languages=["FRE"], sources=None),
    label: str = "umls",
    span_setter: SpanSetterArg = {"ents": True, "umls": True},
):
    """
    The `eds.umls` pipeline component matches the UMLS (Unified Medical Language System
    from NIH) terminology.

    !!! warning "Very low recall"

        When using the `exact` matching mode, this component has a very poor recall
        performance. We can use the `simstring` mode to retrieve approximate matches,
        albeit at the cost of a significantly higher computation time.

    Examples
    --------
    `eds.umls` is an additional module that needs to be setup by:

    1. `pip install -U umls_downloader`
    2. [Signing up](https://uts.nlm.nih.gov/uts/signup-login) for a UMLS Terminology
       Services Account. After filling a short form, you will receive your token API
       within a few days.
    3. Set `UMLS_API_KEY` locally: `export UMLS_API_KEY=your_api_key`

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.umls())

    text = "Grosse toux: le malade a été mordu par des Amphibiens " "sous le genou"

    doc = nlp(text)

    doc.ents
    # Out: (toux, a, par, Amphibiens, genou)

    ent = doc.ents[0]

    ent.label_
    # Out: umls

    ent._.umls
    # Out: C0010200
    ```

    You can easily change the default languages and sources with the `pattern_config`
    argument:

    ```python
    import edsnlp, edsnlp.pipes as eds

    # Enable the French and English languages, through the French MeSH and LOINC
    pattern_config = dict(languages=["FRE", "ENG"], sources=["MSHFRE", "LNC"])

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.umls(pattern_config=pattern_config))
    ```

    See more options of languages and sources
    [here](https://www.nlm.nih.gov/research/umls/sourcereleasedocs/index.html).

    Parameters
    ----------
    nlp : PipelineProtocol
        spaCy `Language` object.
    name : str
        The name of the pipe
    attr : Union[str, Dict[str, str]]
        Attribute to match on, eg `TEXT`, `NORM`, etc.
    ignore_excluded : bool
        Whether to skip excluded tokens during matching.
    ignore_space_tokens : bool
        Whether to skip space tokens during matching.
    term_matcher : TerminologyTermMatcher
        The term matcher to use, either "exact" or "simstring"
    term_matcher_config : Dict[str, Any]
        The configuration for the term matcher
    pattern_config : Dict[str, Any]
        The pattern retriever configuration
    label : str
        Label name to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Authors and citation
    --------------------
    The `eds.umls` pipeline was developed by AP-HP's Data Science team and INRIA
    SODA's team.
    """
    return TerminologyMatcher(
        nlp=nlp,
        name=name,
        regex=dict(),
        terms=get_patterns(pattern_config),
        attr=attr,
        ignore_excluded=ignore_excluded,
        ignore_space_tokens=ignore_space_tokens,
        term_matcher=term_matcher,
        term_matcher_config=term_matcher_config,
        label=label,
        span_setter=span_setter,
    )



from typing import Union

from spacy.tokens import Span

from edsnlp.matchers.utils import get_text

DIGITS_VALUE = list(range(11))
DIGITS_STR = [
    ["zero"],
    ["un", "une", "i"],
    ["deux", "ii"],
    ["trois", "iii"],
    ["quatre", "iv"],
    ["cinq", "v"],
    ["six", "vi"],
    ["sept", "vii"],
    ["huit", "viii"],
    ["neuf", "ix"],
    ["dix", "x"],
]

DIGITS_MAPPINGS = {
    string: digit for digit, strings in enumerate(DIGITS_STR) for string in strings
}


def parse_digit(s: Union[str, Span], **kwargs):
    if isinstance(s, Span):
        string = get_text(
            s,
            attr=kwargs.get("attr", "TEXT"),
            ignore_excluded=kwargs.get("ignore_excluded", True),
        )
    else:
        string = s
    string = string.lower().strip()
    try:
        return int(string)
    except ValueError:
        parsed = DIGITS_MAPPINGS.get(string, None)
        return parsed



from edsnlp.core import registry

from .endlines import EndLinesMatcher

DEFAULT_CONFIG = dict(
    model_path=None,
)

create_component = registry.factory.register(
    "eds.endlines",
    assigns=["doc.ents", "doc.spans"],
    deprecated=["spaces"],
)(EndLinesMatcher)



from edsnlp.core import registry

from .consultation_dates import ConsultationDatesMatcher

DEFAULT_CONFIG = dict(
    consultation_mention=True,
    town_mention=False,
    document_date_mention=False,
    attr="NORM",
    ignore_excluded=False,
    ignore_spacy_tokens=False,
    label="consultation_date",
    span_setter={"ents": True, "consultation_dates": True},
)

create_component = registry.factory.register(
    "eds.consultation_dates",
    assigns=["doc.spans", "doc.ents"],
    deprecated=["consultation_dates"],
)(ConsultationDatesMatcher)



from pydantic import validate_arguments

from edsnlp.pipelines.base import (
    SpanGetterArg,
    SpanSetterArg,
    validate_span_getter,
    validate_span_setter,
)


def test_span_getter():
    assert validate_span_getter("ents") == {"ents": True}
    assert validate_span_getter(["ents"]) == {"ents": True}
    assert validate_span_getter(["ents", "group"]) == {"ents": True, "group": True}
    assert validate_span_getter({"grp": True}) == {"grp": True}
    assert validate_span_getter({"grp": ["a", "b", "c"]}) == {"grp": ["a", "b", "c"]}


def test_span_setter():
    assert validate_span_setter("ents") == {"ents": True}
    assert validate_span_setter(["ents"]) == {"ents": True}
    assert validate_span_setter(["ents", "group"]) == {"ents": True, "group": True}
    assert validate_span_setter({"grp": True}) == {"grp": True}
    assert validate_span_setter({"grp": ["a", "b", "c"]}) == {"grp": ["a", "b", "c"]}


def test_validate_args():
    @validate_arguments
    def my_func(span_getter: SpanGetterArg, span_setter: SpanSetterArg):
        return span_getter, span_setter

    assert my_func("ents", "ents") == ({"ents": True}, {"ents": True})



import warnings
from pathlib import Path

import context  # noqa
import typer

from edsnlp.conjugator import conjugate
from edsnlp.pipelines.qualifiers.hypothesis.patterns import verbs_eds, verbs_hyp
from edsnlp.pipelines.qualifiers.negation.patterns import verbs as neg_verbs
from edsnlp.pipelines.qualifiers.reported_speech.patterns import verbs as rspeech_verbs

warnings.filterwarnings("ignore")


def conjugate_verbs(
    output_path: Path = typer.Argument(
        "edsnlp/resources/verbs.csv.gz", help="Path to the output CSV table."
    )
) -> None:
    """
    Convenience script to automatically conjugate a set of verbs,
    using mlconjug3 library.
    """

    all_verbs = set(neg_verbs + rspeech_verbs + verbs_eds + verbs_hyp)

    typer.echo(f"Conjugating {len(all_verbs)} verbs...")

    df = conjugate(list(all_verbs))

    typer.echo(f"Saving to {output_path}")

    output_path.parent.mkdir(exist_ok=True, parents=True)
    df.to_csv(output_path, index=False)

    typer.echo("Done !")


if __name__ == "__main__":
    typer.run(conjugate_verbs)



from edsnlp.core import registry

from .myocardial_infarction import MyocardialInfarctionMatcher
from .patterns import default_patterns

DEFAULT_CONFIG = dict(
    patterns=default_patterns,
    label="myocardial_infarction",
    span_setter={"ents": True, "myocardial_infarction": True},
)

create_component = registry.factory.register(
    "eds.myocardial_infarction",
    assigns=["doc.ents", "doc.spans"],
)(MyocardialInfarctionMatcher)



from edsnlp.core import registry

from .patterns import quotes_and_apostrophes
from .quotes import QuotesConverter

DEFAULT_CONFIG = dict(
    quotes=quotes_and_apostrophes,
)

create_component = registry.factory.register(
    "eds.quotes",
    assigns=["token.norm"],
    deprecated=["quotes"],
)(QuotesConverter)



# flake8: noqa: F811
import ast
import importlib
import inspect
import os


def lazify():
    def _get_module_paths(file):
        """
        Reads the content of the current file, parses it with ast and store the
        import path for future potential imports. This is useful to only import
        the module that is requested and avoid loading all the modules at once, since
        some of them are quite heavy, or contain dependencies that are not always
        available.

        For instance:
        > from .trainable.span_qualifier.factory import create_component as
        span_qualifier is stored in the cache as:
        > module_paths["span_qualifier"] = "trainable.span_qualifier.factory"

        Returns
        -------
        Dict[str, Tuple[str, str]]
            The absolute path of the current file.
        """
        module_path = os.path.abspath(file)
        with open(module_path, "r") as f:
            module_content = f.read()
        module_ast = ast.parse(module_content)
        module_paths = {}
        for node in module_ast.body:
            # Lookup TYPE_CHECKING
            if not (
                isinstance(node, ast.If)
                and (
                    (
                        isinstance(node.test, ast.Name)
                        and node.test.id == "TYPE_CHECKING"
                    )
                    or (
                        isinstance(node.test, ast.Attribute)
                        and node.test.attr == "TYPE_CHECKING"
                    )
                )
            ):
                continue
            for import_node in node.body:
                if isinstance(import_node, ast.ImportFrom):
                    for name in import_node.names:
                        module_paths[name.asname or name.name] = (
                            import_node.module,
                            name.name,
                        )

        return module_paths

    def __getattr__(name):
        """
        Imports the actual module if it is in the module_paths dict.

        Parameters
        ----------
        name

        Returns
        -------

        """
        if name in module_paths:
            module_path, module_name = module_paths[name]
            result = getattr(
                importlib.__import__(
                    module_path,
                    fromlist=[module_name],
                    globals=module_globals,
                    level=1,
                ),
                module_name,
            )
            module_globals[name] = result
            return result
        raise AttributeError(f"module {__name__} has no attribute {name}")

    def __dir__():
        """
        Returns the list of available modules.

        Returns
        -------
        List[str]
        """
        return __all__

    # Access upper frame
    module_globals = inspect.currentframe().f_back.f_globals

    module_paths = _get_module_paths(module_globals["__file__"])

    __all__ = list(module_paths.keys())

    module_globals.update(
        {
            "__getattr__": __getattr__,
            "__dir__": __dir__,
            "__all__": __all__,
        }
    )



from typing import Dict, List, Union

import mlconjug3
import pandas as pd


def conjugate_verb(
    verb: str,
    conjugator: mlconjug3.Conjugator,
) -> pd.DataFrame:
    """
    Conjugates the verb using an instance of mlconjug3,
    and formats the results in a pandas `DataFrame`.

    Parameters
    ----------
    verb : str
        Verb to conjugate.
    conjugator : mlconjug3.Conjugator
        mlconjug3 instance for conjugating.

    Returns
    -------
    pd.DataFrame
        Normalized dataframe containing all conjugated forms
        for the verb.
    """

    df = pd.DataFrame(
        conjugator.conjugate(verb).iterate(),
        columns=["mode", "tense", "person", "term"],
    )

    df.term = df.term.fillna(df.person)
    df.loc[df.person == df.term, "person"] = None

    df.insert(0, "verb", verb)

    return df


def conjugate(
    verbs: Union[str, List[str]],
    language: str = "fr",
) -> pd.DataFrame:
    """
    Conjugate a list of verbs.

    Parameters
    ----------
    verbs : Union[str, List[str]]
        List of verbs to conjugate
    language: str
        Language to conjugate. Defaults to French (`fr`).

    Returns
    -------
    pd.DataFrame
        Dataframe containing the conjugations for the provided verbs.
        Columns: `verb`, `mode`, `tense`, `person`, `term`
    """
    if isinstance(verbs, str):
        verbs = [verbs]

    conjugator = mlconjug3.Conjugator(language=language)

    df = pd.concat([conjugate_verb(verb, conjugator=conjugator) for verb in verbs])

    df = df.reset_index(drop=True)

    return df


def get_conjugated_verbs(
    verbs: Union[str, List[str]],
    matches: Union[List[Dict[str, str]], Dict[str, str]],
    language: str = "fr",
) -> List[str]:
    """
    Get a list of conjugated verbs.

    Parameters
    ----------
    verbs : Union[str, List[str]]
        List of verbs to conjugate.
    matches : Union[List[Dict[str, str]], Dict[str, str]]
        List of dictionary describing the mode/tense/persons to keep.
    language : str, optional
        [description], by default "fr" (French)

    Returns
    -------
    List[str]
        List of terms to look for.

    Examples
    --------
    >>> get_conjugated_verbs(
            "aimer",
            dict(mode="Indicatif", tense="Présent", person="1p"),
        )
    ['aimons']
    """

    if isinstance(matches, dict):
        matches = [matches]

    terms = []

    df = conjugate(
        verbs=verbs,
        language=language,
    )

    for match in matches:
        q = " & ".join([f'{k} == "{v}"' for k, v in match.items()])
        terms.extend(df.query(q).term.unique())

    return list(set(terms))



from __future__ import annotations

from typing import (
    Any,
    Dict,
    Optional,
    Sequence,
)

import torch
from spacy.tokens import Doc, Span
from typing_extensions import Literal, TypedDict

from edsnlp.core.pipeline import Pipeline
from edsnlp.core.torch_component import BatchInput, BatchOutput
from edsnlp.pipes.base import BaseComponent
from edsnlp.pipes.trainable.embeddings.typing import (
    SpanEmbeddingComponent,
    WordEmbeddingComponent,
)
from edsnlp.utils.filter import align_spans
from edsnlp.utils.span_getters import SpanGetterArg, get_spans

SpanPoolerBatchInput = TypedDict(
    "SpanPoolerBatchInput",
    {
        "embedding": BatchInput,
        "begins": torch.Tensor,
        "ends": torch.Tensor,
        "sequence_idx": torch.Tensor,
    },
)
"""
embeds: torch.FloatTensor
    Token embeddings to predict the tags from
begins: torch.LongTensor
    Begin offsets of the spans
ends: torch.LongTensor
    End offsets of the spans
sequence_idx: torch.LongTensor
    Sequence (cf Embedding spans) index of the spans
"""


class SpanPooler(SpanEmbeddingComponent, BaseComponent):
    """
    The `eds.span_pooler` component is a trainable span embedding component. It
    generates span embeddings from a word embedding component and a span getter. It can
    be used to train a span classifier, as in `eds.span_classifier`.

    Parameters
    ----------
    nlp: PipelineProtocol
        The pipeline object
    name: str
        Name of the component
    embedding : WordEmbeddingComponent
        The word embedding component
    span_getter: SpanGetterArg
        How to extract the candidate spans and the qualifiers
        to predict or train on.
    pooling_mode: Literal["max", "sum", "mean"]
        How word embeddings are aggregated into a single embedding per span.
    hidden_size : Optional[int]
        The size of the hidden layer. If None, no projection is done and the output
        of the span pooler is used directly.
    """

    def __init__(
        self,
        nlp: Optional[Pipeline] = None,
        name: str = "span_pooler",
        *,
        embedding: WordEmbeddingComponent,
        span_getter: SpanGetterArg,
        pooling_mode: Literal["max", "sum", "mean"] = "mean",
        hidden_size: Optional[int] = None,
    ):
        self.qualifiers = None
        self.output_size = embedding.output_size if hidden_size is None else hidden_size

        super().__init__(nlp, name)

        self.pooling_mode = pooling_mode
        self.span_getter = span_getter
        self.embedding = embedding
        self.projector = (
            torch.nn.Linear(self.embedding.output_size, hidden_size)
            if hidden_size is not None
            else torch.nn.Identity()
        )

    def feed_forward(self, span_embeds: torch.Tensor) -> torch.Tensor:
        return self.projector(span_embeds)

    def set_extensions(self):
        super().set_extensions()

        for qlf in self.qualifiers or ():
            if not Span.has_extension(qlf):
                Span.set_extension(qlf, default=None)

    def preprocess(self, doc: Doc) -> Dict[str, Any]:
        spans = list(get_spans(doc, self.span_getter))
        embedded_spans = list(get_spans(doc, self.embedding.span_getter))
        sequence_idx = []
        begins = []
        ends = []

        embedded_spans_to_idx = {span: i for i, span in enumerate(embedded_spans)}
        for i, (span, embedding_spans) in enumerate(
            zip(spans, align_spans(embedded_spans, spans, sort_by_overlap=True))
        ):
            if (
                len(embedding_spans) == 0
                or embedding_spans[0].start > span.start
                or embedding_spans[0].end < span.end
            ):
                raise Exception(
                    f"Span {span.text!r} is not included in at least one embedding "
                    f"span: {[s.text for s in embedding_spans]}"
                )
            start = embedding_spans[0].start
            sequence_idx.append(embedded_spans_to_idx[embedding_spans[0]])
            begins.append(span.start - start)
            ends.append(span.end - start)
        return {
            "embedding": self.embedding.preprocess(doc),
            "begins": begins,
            "ends": ends,
            "sequence_idx": sequence_idx,
            "num_sequences": len(embedded_spans),
            "$spans": spans,
            "$embedded_spans": embedded_spans,
        }

    def collate(self, batch: Dict[str, Sequence[Any]]) -> SpanPoolerBatchInput:
        sequence_idx = []
        offset = 0
        for indices, seq_length in zip(batch["sequence_idx"], batch["num_sequences"]):
            sequence_idx.extend([offset + idx for idx in indices])
            offset += seq_length

        collated: SpanPoolerBatchInput = {
            "embedding": self.embedding.collate(batch["embedding"]),
            "begins": torch.as_tensor([b for x in batch["begins"] for b in x]),
            "ends": torch.as_tensor([e for x in batch["ends"] for e in x]),
            "sequence_idx": torch.as_tensor(sequence_idx),
        }
        return collated

    # noinspection SpellCheckingInspection
    def forward(self, batch: SpanPoolerBatchInput) -> BatchOutput:
        """
        Apply the span classifier module to the document embeddings and given spans to:
        - compute the loss
        - and/or predict the labels of spans
        If labels are predicted, they are assigned to the `additional_outputs`
        dictionary.

        Parameters
        ----------
        batch: SpanPoolerBatchInput
            The input batch

        Returns
        -------
        BatchOutput
        """
        device = next(self.parameters()).device
        if len(batch["begins"]) == 0:
            return {
                "embeddings": torch.empty(0, self.output_size, device=device),
            }

        embeds = self.embedding.module_forward(batch["embedding"])["embeddings"]
        n_samples, n_words, dim = embeds.shape
        device = embeds.device

        flat_begins = n_words * batch["sequence_idx"] + batch["begins"]
        flat_ends = n_words * batch["sequence_idx"] + batch["ends"]
        flat_embeds = embeds.view(-1, dim)
        flat_indices = torch.cat(
            [
                torch.arange(b, e, device=device)
                for b, e in zip(flat_begins.cpu().tolist(), flat_ends.cpu().tolist())
            ]
        ).to(device)
        offsets = (flat_ends - flat_begins).cumsum(0).roll(1)
        offsets[0] = 0
        span_embeds = torch.nn.functional.embedding_bag(  # type: ignore
            input=flat_indices,
            weight=flat_embeds,
            offsets=offsets,
            mode=self.pooling_mode,
        )
        span_embeds = self.feed_forward(span_embeds)

        return {
            "embeddings": span_embeds,
        }



from typing import Any

import pandas as pd
import streamlit as st
from spacy import displacy

import edsnlp
import edsnlp.pipes as eds
from edsnlp.utils.filter import filter_spans

DEFAULT_TEXT = """\
Motif :
Le patient est admis le 29 août pour des difficultés respiratoires.

Antécédents familiaux :
Le père du patient n'est pas asthmatique.

HISTOIRE DE LA MALADIE
Le patient dit avoir de la toux depuis trois jours. \
Elle a empiré jusqu'à nécessiter un passage aux urgences.
A noter deux petits kystes bénins de 1 et 2cm biopsiés en 2005.

Priorité: 2 (établie par l'IAO à l'entrée)

adicaps ABCD0A12 et ABCD0A13

Conclusion
Possible infection au coronavirus. Prescription de paracétomol pour la fièvre.\
"""

REGEX = """
# RegEx and terms matcher
nlp.add_pipe(
    eds.matcher(
        regex=dict(custom=r"{custom_regex}"),
        attr="NORM",
    ),
)
"""

CODE = """
import edsnlp, edsnlp.pipes as eds

# Declare the pipeline
nlp = edsnlp.blank("eds")

# General-purpose components
nlp.add_pipe(eds.normalizer())
nlp.add_pipe(eds.sentences())
{pipes}
# Qualifier pipes
nlp.add_pipe(eds.negation())
nlp.add_pipe(eds.family())
nlp.add_pipe(eds.hypothesis())
nlp.add_pipe(eds.rspeech())

# Define the note text
text = {text}

# Apply the pipeline
doc = nlp(text)

# Explore matched elements
doc.ents
"""

PIPES = {
    "Drugs": "drugs",
    "CIM10": "cim10",
    "Dates": "dates",
    "Measurements": "measurements",
    "Charlson": "charlson",
    "SOFA": "sofa",
    "Elston & Ellis": "elston_ellis",
    "TNM": "tnm",
    "Priority": "emergency_priority",
    "CCMU": "emergency_ccmu",
    "GEMSA": "emergency_gemsa",
    "Covid": "covid",
    "Adicap": "adicap",
    "Diabetes": "diabetes",
    "Tobacco": "tobacco",
    "AIDS": "aids",
    "Lymphoma": "lymphoma",
    "Leukemia": "leukemia",
    "Solid Tumor": "solid_tumor",
    "CKD": "ckd",
    "Hemiplegia": "hemiplegia",
    "Liver Disease": "liver_disease",
    "Peptic Ulcer Disease": "peptic_ulcer_disease",
    "Connective Tissue Disease": "connective_tissue_disease",
    "COPD": "copd",
    "Dementia": "dementia",
    "Cerebrovascular Accident": "cerebrovascular_accident",
    "Peripheral Vascular Disease": "peripheral_vascular_disease",
    "Congestive Heart Failure": "congestive_heart_failure",
    "Myocardial Infarction": "myocardial_infarction",
    "Alcohol": "alcohol",
}


@st.cache_resource()
def load_model(custom_regex: str, **enabled):
    pipes = []

    # Declare the pipeline
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.sentences())

    for title, name in PIPES.items():
        if name == "drugs":
            if enabled["drugs"]:
                if enabled["fuzzy_drugs"]:
                    nlp.add_pipe(eds.drugs(term_matcher="simstring"))
                    pipes.append('nlp.add_pipe(eds.drugs(term_matcher="simstring"))')
                else:
                    nlp.add_pipe(eds.drugs())
                    pipes.append("nlp.add_pipe(eds.drugs())")

        else:
            if enabled[name]:
                nlp.add_pipe(f"eds.{name}")
                pipes.append(f"nlp.add_pipe(eds.{name}())")

    if pipes:
        pipes.insert(0, "# Entity extraction pipes")

    if custom_regex:
        nlp.add_pipe(
            eds.matcher(
                regex=dict(custom=custom_regex),
                attr="NORM",
            ),
        )

        regex = REGEX.format(custom_regex=custom_regex)

    else:
        regex = ""

    nlp.add_pipe(eds.negation())
    nlp.add_pipe(eds.family())
    nlp.add_pipe(eds.hypothesis())
    nlp.add_pipe(eds.rspeech())

    return nlp, pipes, regex


st.set_page_config(
    page_title="EDS-NLP Demo",
    page_icon="📄",
)

st.title("EDS-NLP")

st.warning(
    "You should **not** put sensitive data in the example, as this application "
    "**is not secure**."
)

st.sidebar.header("About")
st.sidebar.markdown(
    "EDS-NLP is a contributive effort maintained by AP-HP's Data Science team. "
    "Have a look at the "
    "[documentation](https://aphp.github.io/edsnlp/) for "
    "more information on the available components."
)

st.sidebar.header("Pipeline")
st.sidebar.markdown(
    "This example runs a simplistic pipeline detecting a few synonyms for "
    "COVID-related entities.\n\n"
    "You can add or remove pre-defined pipeline components, and see how "
    "the pipeline reacts. You can also search for your own custom RegEx."
)

st.sidebar.header("Custom RegEx")
st_custom_regex = st.sidebar.text_input(
    "Regular Expression:",
    r"asthmatique|difficult[ée]s?\srespiratoires?",
)
st.sidebar.markdown("The RegEx you defined above is detected under the `custom` label.")

st.sidebar.subheader("Pipeline Components")
st_pipes = {}

st_pipes["cim10"] = st.sidebar.checkbox("CIM10 (loading can be slow)", value=False)
st_drugs_container = st.sidebar.columns([1, 2])
st_pipes["drugs"] = st_drugs_container[0].checkbox("Drugs", value=True)
st_fuzzy_drugs = st_drugs_container[1].checkbox(
    "Fuzzy drugs search", value=True, disabled=not st_pipes["drugs"]
)
for title, name in PIPES.items():
    if name == "drugs" or name == "cim10":
        continue
    st_pipes[name] = st.sidebar.checkbox(title, value=True)
st.sidebar.markdown(
    "These are just a few of the components provided out-of-the-box by EDS-NLP. "
    "See the [documentation](https://aphp.github.io/edsnlp/latest/pipes/) "
    "for detail."
)

model_load_state = st.info("Loading model...")

nlp, pipes, regex = load_model(
    fuzzy_drugs=st_fuzzy_drugs,
    custom_regex=st_custom_regex,
    **st_pipes,
)

model_load_state.empty()

st.header("Enter a text to analyse:")
text = st.text_area(
    "Modify the following text and see the pipeline react :",
    DEFAULT_TEXT,
    height=375,
)

doc = nlp(text)
doc.ents = filter_spans(
    (*doc.ents, *doc.spans.get("dates", []), *doc.spans.get("measurements", []))
)

st.header("Visualisation")

st.markdown(
    "The pipeline extracts simple entities using a dictionnary of RegEx (see the "
    "[Export the pipeline section](#export-the-pipeline) for more information)."
)

category20 = [
    "#1f77b4",
    "#aec7e8",
    "#ff7f0e",
    "#ffbb78",
    "#2ca02c",
    "#98df8a",
    "#d62728",
    "#ff9896",
    "#9467bd",
    "#c5b0d5",
    "#8c564b",
    "#c49c94",
    "#e377c2",
    "#f7b6d2",
    "#7f7f7f",
    "#c7c7c7",
    "#bcbd22",
    "#dbdb8d",
    "#17becf",
    "#9edae5",
]

labels = [
    "date",
    "covid",
    "drug",
    "cim10",
    "emergency_priority",
    "sofa",
    "charlson",
    "size",
    "weight",
    "adicap",
]

colors = {label: cat for label, cat in zip(labels, category20)}
colors["custom"] = "linear-gradient(90deg, #aa9cfc, #fc9ce7)"
options = {
    "colors": colors,
}

html = displacy.render(doc, style="ent", options=options)
html = html.replace("line-height: 2.5;", "line-height: 2.25;")
html = (
    '<div style="padding: 10px; border: solid 2px; border-radius: 10px; '
    f'border-color: #afc6e0;">{html}</div>'
)
st.write(html, unsafe_allow_html=True)

data = []
for ent in doc.ents:
    d = dict(
        start=ent.start_char,
        end=ent.end_char,
        text=ent.text,
        label=ent.label_,
        normalized_value=str(ent._.value or ""),
        negation="YES" if ent._.negation else "NO",
        family="YES" if ent._.family else "NO",
        hypothesis="YES" if ent._.hypothesis else "NO",
        reported_speech="YES" if ent._.reported_speech else "NO",
    )

    data.append(d)

st.header("Entity qualification")


def color_qualifiers(val: Any) -> str:
    """
    Add color to qualifiers.

    Parameters
    ----------
    val : Any
        DataFrame value

    Returns
    -------
    str
        style
    """
    if val == "NO":
        return "color: #dc3545;"
    elif val == "YES":
        return "color: #198754;"
    return ""


if data:
    df = pd.DataFrame.from_records(data)
    df.normalized_value = df.normalized_value.replace({"None": ""})

    df = df.style.applymap(color_qualifiers)

    st.dataframe(df)

else:
    st.markdown("You pipeline did not match any entity...")

pipes_text = ""

if pipes:
    pipes_text += "\n" + "\n".join(pipes) + "\n"
if regex:
    pipes_text += regex

code = CODE.format(
    pipes=pipes_text,
    text=f'"""\n{text}\n"""',
)

st.header("Export the pipeline")
st.markdown(
    "The code below recreates the pipeline. Copy and paste it "
    "in a Jupyter Notebook to interact with it."
)
with st.expander("Show the runnable code"):
    st.markdown(f"```python\n{code}\n```\n\nThis code runs as is.")



COMPLICATIONS = [
    r"nephropat",
    r"neuropat",
    r"retinopat",
    r"glomerulopathi",
    r"glomeruloscleros",
    r"angiopathi",
    r"origine",
]

main_pattern = dict(
    source="main",
    regex=[
        r"\bds?n?id\b",
        r"\bdiabet[^o]",
        r"\bdb\b",
        r"\bdt.?(i|ii|1|2)\b",
    ],
    exclude=dict(
        regex=[
            "insipide",
            "nephrogenique",
            "aigu",
            r"\bdr\b",  # Dr. ...
            "endocrino",  # Section title
            "soins aux pieds",  # Section title
            "nutrition",  # Section title
            r"\s?:\n+\W+(?!oui|non|\W)",  # General pattern for section title
        ],
        window=(-5, 5),
    ),
    regex_attr="NORM",
    assign=[
        dict(
            name="complicated_before",
            regex=r"(" + r"|".join(COMPLICATIONS + ["origine"]) + r")",
            window=-3,
        ),
        dict(
            name="complicated_after",
            regex=r"("
            + r"|".join([r"(?<!sans )compli", r"(?<!a)symptomatique"] + COMPLICATIONS)
            + r")",
            window=12,
        ),
        dict(
            name="type",
            regex=r"type.(i|ii|1|2)",
            window=6,
        ),
        dict(
            name="insulin",
            regex=r"insulino.?(dep|req)",
            window=6,
        ),
        dict(
            name="corticoid",
            regex=r"(bctc\b|cortico(?:.?induit)?)",
            window=6,
        ),
    ],
)

complicated_pattern = dict(
    source="complicated",
    regex=[
        r"(mal|maux).perforants?(.plantaire)?",
        r"pieds? diabeti",
    ],
    exclude=dict(
        regex="soins aux",  # Section title
        window=-2,
    ),
    regex_attr="NORM",
)

default_patterns = [
    main_pattern,
    complicated_pattern,
]



import datetime

import pytest
import pytz
import spacy
from pytest import fixture

from edsnlp.core import PipelineProtocol
from edsnlp.pipelines.misc.dates.models import AbsoluteDate, Relative
from edsnlp.utils.examples import parse_example

TZ = pytz.timezone("Europe/Paris")
TODAY = datetime.date.today()

examples = [
    (
        "Le patient est venu en <ent norm='2019-??-??' year=2019>2019</ent> pour une "
        "consultation"
    ),
    "Le patient est venu <ent norm='-1 day' direction=past day=1>hier</ent>",
    "le <ent norm='2021-09-04' day=4 month=9 year=2021>04/09/2021</ent>",
    (
        "Il est cas contact <ent norm='-1 week' direction=past week=1>"
        "depuis la semaine dernière</ent>"
    ),
    "le <ent norm='????-08-09' day=9 month=8>09/08</ent>",
    "Le patient est venu le <ent norm='????-08-04' day=4 month=8>4 août</ent>",
    (
        "Le patient est venu le <ent norm='????-08-04 11h13m' day=4 month=8 "
        "hour=11 minute=13>4 août à 11h13</ent>"
    ),
    "Il est venu le <ent norm='????-09-01' day=1 month=9>1er Septembre</ent> pour",
    (
        "Il est venu en <ent norm='2020-10-??' month=10 year=2020>octobre 2020</ent> "
        "pour..."
    ),
    (
        "Il est venu <ent norm='-3 months' direction=past month=3>il y a "
        "trois mois</ent> pour..."
    ),
    (
        "Il lui était arrivé la même chose <ent norm='-1 year' "
        "direction=past year=1>il y a un an</ent>."
    ),
    (
        "Il est venu le <ent norm='2001-09-20' day=20 month=9 "
        "year=2001>20/09/2001</ent> pour..."
    ),
    (
        "Consultation du <ent norm='2019-07-03' bound=from "
        "day=3 month=7 year=2019>03 07 19</ent>"
    ),
    "En <ent norm='2017-11-??' month=11 year=2017>11/2017</ent> stabilité sur...",
    "<ent norm='-3 months' direction=past month=3>depuis 3 mois</ent>",
    "- <ent norm='2004-12-??' month=12 year=2004>Décembre 2004</ent> :",
    "- <ent norm='2005-06-??' month=6 year=2005>Juin 2005</ent>:  ",
    # "-<ent norm=" month=6 year=2005>Juin 2005</ent>:  ",  # issues with "fr" language
    "<ent norm='2017-09-??' month=9 year=2017>sept 2017</ent> :",
    (
        "<ent norm='-1 year' direction=past year=1>il y a 1 an</ent> "
        "<ent norm='during 1 month' mode=duration month=1>pdt 1 mois</ent>"
    ),
    (
        "Prélevé le : <ent norm='2016-04-22' day=22 month=4 year=2016>22/04/2016</ent> "
        "\n78 rue du Général Leclerc"
    ),
    "Le <ent norm='????-01-07' day=7 month=1>07/01</ent>.",
    "Il est venu en <ent norm='????-08-??' month=8>août</ent>.",
    "Il est venu <ent norm='~0 day' day=0 direction=current>ce jour</ent>.",
    "CS le <ent norm='2017-01-11' day=11 month=1 year=2017>11-01-2017</ent> 1/3",
    "Vu le <ent norm='2017-01-11' day=11 month=1 year=2017>11 janvier\n2017</ent> .",
]


@fixture(autouse=True)
def add_date_pipeline(blank_nlp: PipelineProtocol):
    blank_nlp.add_pipe("eds.dates", config=dict(detect_periods=True, as_ents=True))


def test_dates_component(blank_nlp: PipelineProtocol):
    note_datetime = datetime.datetime(year=1993, month=9, day=23)

    for example in examples:
        text, entities = parse_example(example)

        doc = blank_nlp(text)
        spans = sorted(doc.spans["dates"] + doc.spans["durations"])

        assert len(spans) == len(entities)
        assert len(doc.ents) == len(entities)

        for span, entity in zip(spans, entities):
            assert span.text == text[entity.start_char : entity.end_char]

            date = span._.date if span.label_ == "date" else span._.duration
            d = {modifier.key: modifier.value for modifier in entity.modifiers}
            norm = d.pop("norm")
            if "direction" in d:
                d["mode"] = "relative"
            if "mode" not in d:
                d["mode"] = "absolute"

            assert date.dict(exclude_none=True) == d
            assert date.norm() == norm

            set_d = set(d)

            d.pop("mode", None)
            d.pop("direction", None)
            d.pop("bound", None)
            if isinstance(date, AbsoluteDate) and {"year", "month", "day"}.issubset(
                set_d
            ):
                assert date.to_datetime() == TZ.localize(datetime.datetime(**d))

            elif isinstance(date, AbsoluteDate):
                assert date.to_datetime() is None

                # no year
                if {"month", "day"}.issubset(set_d) and {"year"}.isdisjoint(set_d):
                    d["year"] = note_datetime.year
                    assert date.to_datetime(
                        note_datetime=note_datetime, infer_from_context=True
                    ) == TZ.localize(datetime.datetime(**d))

                # no day
                if {"month", "year"}.issubset(set_d) and {"day"}.isdisjoint(set_d):
                    d["day"] = 1
                    assert date.to_datetime(
                        note_datetime=note_datetime, infer_from_context=True
                    ) == TZ.localize(datetime.datetime(**d))

                # year only
                if {"year"}.issubset(set_d) and {"day", "month"}.isdisjoint(set_d):
                    d["day"] = 1
                    d["month"] = 1
                    assert date.to_datetime(
                        note_datetime=note_datetime, infer_from_context=True
                    ) == TZ.localize(datetime.datetime(**d))

                # month only
                if {"month"}.issubset(set_d) and {"day", "year"}.isdisjoint(set_d):
                    d["day"] = 1
                    d["year"] = note_datetime.year
                    assert date.to_datetime(
                        note_datetime=note_datetime, infer_from_context=True
                    ) == TZ.localize(datetime.datetime(**d))

            elif isinstance(date, Relative):
                assert date.to_datetime() is None
            else:
                assert date.to_duration()
                assert date.to_datetime(note_datetime=note_datetime)


def test_periods(blank_nlp: PipelineProtocol):
    period_examples = [
        "à partir de <ent>juin 2017 pendant trois semaines</ent>",
        "du <ent>5 juin au 6 juillet</ent>",
    ]

    for example in period_examples:
        text, entities = parse_example(example)

        doc = blank_nlp(text)

        assert len(doc.spans["periods"]) == len(entities)

        for span, entity in zip(doc.spans["periods"], entities):
            assert span.text == text[entity.start_char : entity.end_char]


@pytest.mark.parametrize("with_time", [False, True])
def test_time(with_time: bool):
    nlp = spacy.blank("eds")
    nlp.add_pipe("eds.dates", config={"detect_time": with_time})

    if with_time:
        time_examples = [
            "Vu le <ent norm='2012-01-11 11h34m'>11/01/2012 à 11h34</ent> pour radio.",
        ]
    else:
        time_examples = [
            "Vu le <ent norm='2012-01-11'>11/01/2012</ent> à 11h34 pour radio.",
        ]

    for example in time_examples:
        text, entities = parse_example(example)

        doc = nlp(text)

        spans = sorted(doc.spans["dates"] + doc.spans["durations"])

        assert len(spans) == len(entities)

        for span, entity in zip(spans, entities):
            assert span.text == text[entity.start_char : entity.end_char]
            norm = next(m.value for m in entity.modifiers if m.key == "norm")
            assert span._.date.norm() == norm


def test_false_positives(blank_nlp: PipelineProtocol):
    counter_examples = [
        "page 1/1",  # Often found in the form `1/1` only
        "40 00",
        "06 12 34 56 78",
        "bien mais",
        "thierry",
        "436",
        "12.0-16",
        "27.0-33",
        "7.0-11",
        "03-0.70",
        "4.09-11",
        "2/2CR Urgences PSL",
        "Dextro : 5.7 mmol/l",
        "2.5",
    ]

    for example in counter_examples:
        doc = blank_nlp(example)

        assert len((*doc.spans["dates"], *doc.spans["durations"])) == 0


def test_dates_on_ents_only():
    text = (
        "Le patient est venu hier (le 04/09/2021) pour un test PCR.\n"
        "Il est cas contact <ent>depuis la semaine dernière</ent>, "
        "le <ent>09/08</ent> (<ent>2021-08-09</ent>)."
    )

    nlp = spacy.blank("eds")

    nlp.add_pipe("eds.sentences")
    nlp.add_pipe("eds.matcher", config=dict(terms={"contact": "contact"}))
    nlp.add_pipe("eds.dates", config=dict(on_ents_only=True))

    text, entities = parse_example(text)

    doc = nlp(text)

    assert len(doc.ents) == 1

    spans = sorted(doc.spans["dates"] + doc.spans["durations"])

    assert len(spans) == len(entities)

    for span, entity in zip(spans, entities):
        assert span.text == text[entity.start_char : entity.end_char]


def test_illegal_dates(blank_nlp):
    texts = (
        " Le 31/06/17, la dernière dose.",
        " Le 30/02/18 n'est pas une vraie date",
    )
    for text in texts:
        doc = blank_nlp(text)
        ent = sorted((*doc.spans["dates"], *doc.spans["durations"]))[0]
        assert ent._.date.to_datetime() is None


@pytest.mark.parametrize(
    "dt",
    [
        datetime.datetime.now().strftime("%Y-%m-%d"),
        datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        datetime.datetime.now(),
        datetime.datetime.now().timestamp(),
        datetime.date.today(),
        None,
    ],
)
def test_note_datetime(blank_nlp, dt):
    doc = blank_nlp.make_doc("Empty doc")
    doc._.note_datetime = dt
    if dt is not None:
        assert doc._.note_datetime.year == TODAY.year
        assert doc._.note_datetime.month == TODAY.month
        assert doc._.note_datetime.day == TODAY.day



from edsnlp.core import registry

from .congestive_heart_failure import CongestiveHeartFailureMatcher
from .patterns import default_patterns

DEFAULT_CONFIG = dict(
    patterns=default_patterns,
    label="congestive_heart_failure",
    span_setter={"ents": True, "congestive_heart_failure": True},
)

create_component = registry.factory.register(
    "eds.congestive_heart_failure",
    assigns=["doc.ents", "doc.spans"],
)(CongestiveHeartFailureMatcher)



from typing import Any, Dict

from typing_extensions import Literal

from edsnlp.core import PipelineProtocol, registry
from edsnlp.pipes.core.terminology.terminology import TerminologyMatcher

from ...base import SpanSetterArg
from .patterns import get_patterns

DEFAULT_CONFIG = dict(
    attr="NORM",
    ignore_excluded=False,
    ignore_space_tokens=False,
    term_matcher="exact",
    term_matcher_config={},
    label="cim10",
    span_setter={"ents": True, "cim10": True},
)


@registry.factory.register(
    "eds.cim10",
    assigns=["doc.ents", "doc.spans"],
)
def create_component(
    nlp: PipelineProtocol,
    name: str = "cim10",
    *,
    attr: str = "NORM",
    ignore_excluded: bool = False,
    ignore_space_tokens: bool = False,
    term_matcher: Literal["exact", "simstring"] = "exact",
    term_matcher_config: Dict[str, Any] = {},
    label: str = "cim10",
    span_setter: SpanSetterArg = {"ents": True, "cim10": True},
):
    """
    The `eds.cim10` pipeline component extract terms from documents using the CIM10
    (French-language ICD) terminology as a reference.

    !!! warning "Very low recall"

        When using the `exact` matching mode, this component has a very poor recall
        performance. We can use the `simstring` mode to retrieve approximate matches,
        albeit at the cost of a significantly higher computation time.

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.cim10(term_matcher="simstring"))

    text = "Le patient est suivi pour fièvres typhoïde et paratyphoïde."

    doc = nlp(text)

    doc.ents
    # Out: (fièvres typhoïde et paratyphoïde,)

    ent = doc.ents[0]

    ent.label_
    # Out: cim10

    ent.kb_id_
    # Out: A01
    ```

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : str
        The name of the component
    attr : str
        The default attribute to use for matching.
    ignore_excluded : bool
        Whether to skip excluded tokens (requires an upstream
        pipeline to mark excluded tokens).
    ignore_space_tokens : bool
        Whether to skip space tokens during matching.
    term_matcher: TerminologyTermMatcher
        The matcher to use for matching phrases ?
        One of (exact, simstring)
    term_matcher_config: Dict[str,Any]
        Parameters of the matcher term matcher
    label : str
        Label name to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Returns
    -------
    TerminologyMatcher

    Authors and citation
    --------------------
    The `eds.cim10` pipeline was developed by AP-HP's Data Science team.
    """
    return TerminologyMatcher(
        nlp=nlp,
        name=name,
        regex=dict(),
        terms=get_patterns(),
        attr=attr,
        ignore_excluded=ignore_excluded,
        ignore_space_tokens=ignore_space_tokens,
        term_matcher=term_matcher,
        term_matcher_config=term_matcher_config,
        label=label,
        span_setter=span_setter,
    )



"""
These section titles were extracted from a work performed by Ivan Lerner at AP-HP.
It supplied a number of documents annotated for section titles.

The section titles were reviewed by Gilles Chatellier, who gave meaningful insights.

See sections/section-dataset notebook for detail.
"""

allergies = [r"allergies"]

antecedents = [
    r"antecedents",
    r"antecedents medicaux et chirurgicaux",
    r"antecedents personnels",
    r"antecedents medicaux",
    r"antecedents chirurgicaux",
    r"atcd",
]

antecedents_familiaux = [r"antecedents familiaux"]

traitements_entree = [
    r"attitude therapeutique initiale",
    r"traitement a l'entree",
    r"traitement actuel",
    r"traitement en cours",
    r"traitements a l'entree",
]

conclusion = [
    r"au total",
    r"conclusion",
    r"conclusion de sortie",
    r"syntese medicale / conclusion",
    r"synthese",
    r"synthese medicale",
    r"synthese medicale/conclusion",
    r"conclusion medicale",
]

conclusion_entree = [r"conclusion a l'entree"]

habitus = [
    r"contexte familial et social",
    r"habitus",
    r"mode de vie",
    r"mode de vie - scolarite",
    r"situation sociale, mode de vie",
]

correspondants = [r"correspondants"]

diagnostic = [r"diagnostic retenu"]

donnees_biometriques_entree = [
    r"donnees biometriques et parametres vitaux a l'entree",
    r"parametres vitaux et donnees biometriques a l'entree",
]

examens = [r"examen clinique", r"examen clinique a l'entree"]

examens_complementaires = [
    r"examen(s) complementaire(s)",
    r"examens complementaires",
    r"examens complementaires a l'entree",
    r"examens complementaires realises a l'entree",
    r"examens complementaires realises pendant le sejour",
    r"examens para-cliniques",
    r"imagerie post-operatoire",
]

facteurs_de_risques = [r"facteurs de risque", r"facteurs de risques"]

histoire_de_la_maladie = [
    r"histoire de la maladie",
    r"histoire de la maladie - explorations",
    r"histoire de la maladie actuelle",
    r"histoire du poids",
    r"histoire recente",
    r"histoire recente de la maladie",
    r"rappel clinique",
    r"resume",
    r"resume clinique",
    r"resume clinique - histoire de la maladie",
    r"antecedents et histoire de la maladie",
]

actes = [r"intervention"]

motif = [
    r"motif",
    r"motif d'hospitalisation",
    r"motif de l'hospitalisation",
    r"motif medical",
]

prescriptions = [r"prescriptions de sortie", r"prescriptions medicales de sortie"]

traitements_sortie = [r"traitement de sortie"]

evolution = [r"evolution", r"evolution et examen clinique aux lits portes :"]

modalites_sortie = [r"modalites de sortie", r"devenir du patient"]

vaccinations = [r"vaccinations", r"vaccination"]

introduction = [
    r"compte.?rendu d'hospitalisation.{0,30}",
]

sections = {
    r"allergies": allergies,
    r"antécédents": antecedents,
    r"antécédents familiaux": antecedents_familiaux,
    r"traitements entrée": traitements_entree,
    r"conclusion": conclusion,
    r"conclusion entrée": conclusion_entree,
    r"habitus": habitus,
    r"correspondants": correspondants,
    r"diagnostic": diagnostic,
    r"données biométriques entrée": donnees_biometriques_entree,
    r"examens": examens,
    r"examens complémentaires": examens_complementaires,
    r"facteurs de risques": facteurs_de_risques,
    r"histoire de la maladie": histoire_de_la_maladie,
    r"actes": actes,
    r"motif": motif,
    r"prescriptions": prescriptions,
    r"traitements sortie": traitements_sortie,
    r"evolution": evolution,
    r"modalites sortie": modalites_sortie,
    r"vaccinations": vaccinations,
    r"introduction": introduction,
}



from __future__ import annotations

import pickle
import sys
from typing import Optional

import dill

from edsnlp.core.lazy_collection import LazyCollection
from edsnlp.data.base import BaseWriter
from edsnlp.data.converters import set_current_tokenizer
from edsnlp.data.spark import SparkReader, SparkWriter
from edsnlp.utils.collections import batchify, flatten

try:
    from koalas.dataframe import DataFrame as KoalasDataFrame
except ImportError:  # pragma: no cover
    KoalasDataFrame = None


class Broadcasted:
    def __init__(self, value):
        self.value = value

    def __getstate__(self):
        return dill.dumps(self.value, byref=True, recurse=True)

    def __setstate__(self, state):  # pragma: no cover
        self.value = dill.loads(state)


def execute_spark_backend(
    lc: LazyCollection,
):
    """
    This execution mode uses Spark to parallelize the processing of the documents.
    The documents are first stored in a Spark DataFrame (if it was not already the case)
    and then processed in parallel using Spark.

    Beware, if the original reader was not a SparkReader (`edsnlp.data.from_spark`), the
    *local docs* → *spark dataframe* conversion might take some time, and the whole
    process might be slower than using the `multiprocessing` backend.
    """
    import pyspark.sql.types as T
    from pyspark.sql import SparkSession

    from edsnlp.utils.spark_dtypes import schema_warning, spark_interpret_dicts_as_rows

    try:
        getActiveSession = SparkSession.getActiveSession
    except AttributeError:

        def getActiveSession() -> Optional["SparkSession"]:  # pragma: no cover
            from pyspark import SparkContext

            sc = SparkContext._active_spark_context
            if sc is None:
                return None
            else:
                assert sc._jvm is not None
                if sc._jvm.SparkSession.getActiveSession().isDefined():
                    SparkSession(sc, sc._jvm.SparkSession.getActiveSession().get())
                    try:
                        return SparkSession._activeSession
                    except AttributeError:
                        try:
                            return SparkSession._instantiatedSession
                        except AttributeError:
                            return None
                else:
                    return None

    # Get current spark session
    spark = getActiveSession() or SparkSession.builder.getOrCreate()

    reader = lc.reader
    writer = lc.writer

    if (
        writer is not None
        # check if finalize has been overridden
        and writer.finalize.__func__ is not BaseWriter.finalize
    ):
        raise ValueError(
            "The Spark backend does not support writers that need to be finalized."
        )

    if isinstance(lc.reader, SparkReader):
        df = lc.reader.data
    else:
        with spark_interpret_dicts_as_rows():
            df = spark.sparkContext.parallelize(
                [
                    {"content": pickle.dumps(item, -1)}
                    for item, count in reader.read_main()
                ]
            ).toDF(T.StructType([T.StructField("content", T.BinaryType())]))

    def process_partition(iterator):  # pragma: no cover
        lc: LazyCollection = bc.value
        try:
            sys.modules["torch"].set_grad_enabled(False)
        except (AttributeError, KeyError):
            pass

        results = []
        tasks = (
            # Maybe we should skip read_worker here ? it's a no-op for SparkReader
            lc.reader.read_worker(iterator)
            if isinstance(lc.reader, SparkReader)
            else (
                task
                for row in iterator
                for task in lc.reader.read_worker([pickle.loads(row["content"])])
            )
        )
        for batch in batchify(
            tasks,
            batch_size=lc.batch_size,
        ):
            with lc.cache():
                for name, pipe, kwargs, tokenizer in lc.pipeline:
                    with set_current_tokenizer(tokenizer):
                        if hasattr(pipe, "batch_process"):
                            batch = pipe.batch_process(batch, **kwargs)
                        else:
                            batch = [pipe(doc, **kwargs) for doc in batch]  # type: ignore
            results.extend(batch)

        if lc.writer:
            results, count = lc.writer.write_worker(results)

        if isinstance(lc.writer, SparkWriter):
            return results
        else:
            return [{"content": pickle.dumps(results, -1)}]

    with lc.eval():
        bc = Broadcasted(lc.worker_copy())

        if isinstance(writer, SparkWriter):
            rdd = df.rdd.mapPartitions(process_partition)
            with spark_interpret_dicts_as_rows():
                results = spark.createDataFrame(rdd, schema=writer.dtypes)

            if writer.dtypes is None and writer.show_dtypes:
                schema_warning(results.schema)
            return results

        results = (
            pickle.loads(item["content"])
            for item in df.rdd.mapPartitions(process_partition).toLocalIterator()
        )
        return flatten(results) if writer is None else writer.write_main(results)



from edsnlp.utils.regex import make_pattern

letter_days = [
    r"(?P<day_01>premier|1\s*er)",
    r"(?P<day_02>deux)",
    r"(?P<day_03>trois)",
    r"(?P<day_04>quatre)",
    r"(?P<day_05>cinq)",
    r"(?P<day_06>six)",
    r"(?P<day_07>sept)",
    r"(?P<day_08>huit)",
    r"(?P<day_09>neuf)",
    r"(?P<day_10>dix)",
    r"(?P<day_11>onze)",
    r"(?P<day_12>douze)",
    r"(?P<day_13>treize)",
    r"(?P<day_14>quatorze)",
    r"(?P<day_15>quinze)",
    r"(?P<day_16>seize)",
    r"(?P<day_17>dix\-?\s*sept)",
    r"(?P<day_18>dix\-?\s*huit)",
    r"(?P<day_19>dix\-?\s*neuf)",
    r"(?P<day_20>vingt)",
    r"(?P<day_21>vingt\-?\s*et\-?\s*un)",
    r"(?P<day_22>vingt\-?\s*deux)",
    r"(?P<day_23>vingt\-?\s*trois)",
    r"(?P<day_24>vingt\-?\s*quatre)",
    r"(?P<day_25>vingt\-?\s*cinq)",
    r"(?P<day_26>vingt\-?\s*six)",
    r"(?P<day_27>vingt\-?\s*sept)",
    r"(?P<day_28>vingt\-?\s*huit)",
    r"(?P<day_29>vingt\-?\s*neuf)",
    r"(?P<day_30>trente)",
    r"(?P<day_31>trente\-?\s*et\-?\s*un)",
]

letter_day_pattern = make_pattern(letter_days)

numeric_day_pattern = r"(?<!\d)(0?[1-9]|[12]\d|3[01])(?!\d)"
lz_numeric_day_pattern = r"(?<!\d)(0[1-9]|[12]\d|3[01])(?!\d)"
nlz_numeric_day_pattern = r"(?<!\d)([1-9]|[12]\d|3[01])(?!\d)"

numeric_day_pattern = f"(?P<day>{numeric_day_pattern})"
lz_numeric_day_pattern = f"(?P<day>{lz_numeric_day_pattern})"

day_pattern = f"({letter_day_pattern}|{numeric_day_pattern})"



import os
from io import BytesIO

import pytest
from confit import Config
from confit.errors import ConfitValidationError
from confit.registry import validate_arguments
from spacy.tokens import Doc

import edsnlp
import edsnlp.pipes as eds
from edsnlp import Pipeline, registry
from edsnlp.core.registries import CurriedFactory
from edsnlp.pipes.base import BaseComponent


class CustomClass:
    pass

    def __call__(self, doc: Doc) -> Doc:
        return doc


def test_add_pipe_factory():
    model = edsnlp.blank("eds")
    model.add_pipe("eds.normalizer", name="normalizer")
    assert "normalizer" in model.pipe_names
    assert model.has_pipe("normalizer")

    model.add_pipe("eds.sentences", name="sentences")
    assert "sentences" in model.pipe_names
    assert model.has_pipe("sentences")

    with pytest.raises(ValueError):
        model.get_pipe("missing-pipe")


def test_add_pipe_component():
    model = edsnlp.blank("eds")
    model.add_pipe(eds.normalizer(nlp=model), name="normalizer")
    assert "normalizer" in model.pipe_names
    assert model.has_pipe("normalizer")
    assert model.pipes.normalizer is model.get_pipe("normalizer")

    model.add_pipe(eds.sentences(nlp=model), name="sentences")
    assert "sentences" in model.pipe_names
    assert model.has_pipe("sentences")
    assert model.pipes.sentences is model.get_pipe("sentences")

    with pytest.raises(ValueError):
        model.add_pipe(
            eds.sentences(nlp=model, name="sentences"),
            config={"punct_chars": ".?!"},
        )

    with pytest.raises(ValueError):
        model.add_pipe(CustomClass())


def test_sequence(frozen_ml_nlp: Pipeline):
    assert len(frozen_ml_nlp.pipeline) == 3
    assert list(frozen_ml_nlp.pipeline) == [
        ("sentences", frozen_ml_nlp.get_pipe("sentences")),
        ("transformer", frozen_ml_nlp.get_pipe("transformer")),
        ("ner", frozen_ml_nlp.get_pipe("ner")),
    ]
    assert list(frozen_ml_nlp.torch_components()) == [
        ("transformer", frozen_ml_nlp.get_pipe("transformer")),
        ("ner", frozen_ml_nlp.get_pipe("ner")),
    ]


def test_disk_serialization(tmp_path, ml_nlp):
    nlp = ml_nlp

    assert nlp.get_pipe("transformer").stride == 96
    ner = nlp.get_pipe("ner")
    ner.update_labels(["PERSON", "GIFT"])

    os.makedirs(tmp_path / "model", exist_ok=True)
    nlp.to_disk(tmp_path / "model")

    print("tmp_path", tmp_path, list((tmp_path / "model/transformer").iterdir()))

    assert (tmp_path / "model" / "config.cfg").exists()
    assert (tmp_path / "model" / "ner" / "parameters.safetensors").exists()
    assert (tmp_path / "model" / "transformer" / "parameters.safetensors").exists()
    # fmt: off
    assert (
          (tmp_path / "model" / "transformer" / "pytorch_model.bin").exists() or
          (tmp_path / "model" / "transformer" / "model.safetensors").exists()
    )
    # fmt: on

    assert (tmp_path / "model" / "config.cfg").read_text() == (
        config_str.replace("components = ${components}\n", "").replace(
            "prajjwal1/bert-tiny", "./transformer"
        )
    )

    nlp = edsnlp.load(
        tmp_path / "model",
        overrides={"components": {"transformer": {"stride": 64}}},
    )
    assert nlp.get_pipe("ner").labels == ["PERSON", "GIFT"]
    assert nlp.get_pipe("transformer").stride == 64


config_str = """\
[nlp]
lang = "eds"
pipeline = ["sentences", "transformer", "ner"]
components = ${components}

[nlp.tokenizer]
@tokenizers = "eds.tokenizer"

[components]

[components.sentences]
@factory = "eds.sentences"

[components.transformer]
@factory = "eds.transformer"
model = "prajjwal1/bert-tiny"
window = 128
stride = 96

[components.ner]
@factory = "eds.ner_crf"
embedding = ${components.transformer}
target_span_getter = ["ents", "ner-preds"]
mode = "independent"
labels = ["PERSON", "GIFT"]
infer_span_setter = false
window = 40
stride = 20

[components.ner.span_setter]
ents = true

"""


def test_validate_config():
    @validate_arguments
    def function(model: Pipeline):
        print(model.pipe_names)
        assert len(model.pipe_names) == 3

    function(Config.from_str(config_str).resolve(registry=registry)["nlp"])


def test_torch_module(frozen_ml_nlp: Pipeline):
    with frozen_ml_nlp.train(True):
        for name, component in frozen_ml_nlp.torch_components():
            assert component.training is True

    with frozen_ml_nlp.train(False):
        for name, component in frozen_ml_nlp.torch_components():
            assert component.training is False

    frozen_ml_nlp.to("cpu")


def test_cache(frozen_ml_nlp: Pipeline):
    from edsnlp.core.torch_component import _caches

    text = "Ceci est un exemple"
    frozen_ml_nlp(text)

    doc = frozen_ml_nlp.make_doc(text)
    with frozen_ml_nlp.cache():
        for name, pipe in frozen_ml_nlp.pipeline:
            # This is a hack to get around the ambiguity
            # between the __call__ method of Pytorch modules
            # and the __call__ methods of spacy components
            if hasattr(pipe, "batch_process"):
                doc = next(iter(pipe.batch_process([doc])))
            else:
                doc = pipe(doc)
        trf_forward_cache_entries = [
            key
            for key in _caches["default"]
            if isinstance(key, tuple) and key[0] == "forward"
        ]
        assert len(trf_forward_cache_entries) == 2

    assert len(_caches) == 0


def test_select_pipes(frozen_ml_nlp: Pipeline):
    text = "Ceci est un exemple"
    with frozen_ml_nlp.select_pipes(enable=["transformer", "ner"]):
        assert len(frozen_ml_nlp.disabled) == 1
        assert not frozen_ml_nlp(text).has_annotation("SENT_START")
    assert len(frozen_ml_nlp.disabled) == 0


@pytest.mark.skip(reason="Deprecated behavior")
def test_different_names():
    nlp = edsnlp.blank("eds")

    extractor = eds.sentences(nlp=nlp, name="custom_name")

    with pytest.raises(ValueError) as exc_info:
        nlp.add_pipe(extractor, name="sentences")

    assert (
        "The provided name 'sentences' does not "
        "match the name of the component 'custom_name'."
    ) in str(exc_info.value)


def test_load_config(run_in_test_dir):
    nlp = edsnlp.load("training/qlf_config.cfg")
    assert nlp.pipe_names == [
        "normalizer",
        "sentencizer",
        "embedding",
        "covid",
        "qualifier",
    ]


fail_config = """
[nlp]
lang = "eds"
pipeline = ["transformer", "ner"]

[nlp.tokenizer]
@tokenizers = "eds.tokenizer"

[components]

[components.transformer]
@factory = "eds.transformer"
model = "prajjwal1/bert-tiny"
window = 128
stride = 96

[components.ner]
@factory = "eds.ner_crf"
embedding = ${components.transformer}
mode = "error-mode"
span_setter = "ents"
"""


def test_config_validation_error():
    with pytest.raises(ConfitValidationError) as e:
        Pipeline.from_config(Config.from_str(fail_config))

    assert "1 validation error for edsnlp.core.pipeline.Pipeline()" in str(e.value)
    assert "got 'error-mode'" in str(e.value)


def test_add_pipe_validation_error():
    model = edsnlp.blank("eds")
    with pytest.raises(ConfitValidationError) as e:
        model.add_pipe("eds.covid", name="extractor", config={"foo": "bar"})

    assert str(e.value) == (
        "1 validation error for "
        "edsnlp.pipes.ner.covid.factory.create_component()\n"
        "-> extractor.foo\n"
        "   unexpected keyword argument"
    )


def test_spacy_component():
    nlp = edsnlp.blank("fr")
    nlp.add_pipe("sentencizer")


def test_rule_based_pipeline():
    nlp = edsnlp.blank("eds")
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe("eds.covid")

    assert nlp.pipe_names == ["normalizer", "covid"]
    assert nlp.get_pipe("normalizer") == nlp.pipeline[0][1]
    assert nlp.has_pipe("covid")

    with pytest.raises(ValueError) as exc_info:
        nlp.get_pipe("unknown")

    assert str(exc_info.value) == "Pipe 'unknown' not found in pipeline."

    doc = nlp.make_doc("Mon patient a le covid")

    new_doc = nlp(doc)

    assert len(doc.ents) == 1
    assert new_doc is doc

    assert nlp.get_pipe_meta("covid").assigns == ["doc.ents", "doc.spans"]


def test_torch_save(ml_nlp):
    import torch

    ml_nlp.get_pipe("ner").update_labels(["LOC", "PER"])
    buffer = BytesIO()
    torch.save(ml_nlp, buffer)
    buffer.seek(0)
    nlp = torch.load(buffer)
    assert nlp.get_pipe("ner").labels == ["LOC", "PER"]
    assert len(list(nlp("Une phrase. Deux phrases.").sents)) == 2


def test_parameters(frozen_ml_nlp):
    assert len(list(frozen_ml_nlp.parameters())) == 42


def test_missing_factory(nlp):
    with pytest.raises(ValueError) as exc_info:
        nlp.add_pipe("__test_missing_pipe__")

    assert "__test_missing_pipe__" in str(exc_info.value)


@edsnlp.registry.factory("custom-curry-test")
class CustomComponent(BaseComponent):
    def __init__(self, nlp, name):
        self.nlp = nlp

    def __call__(self, doc):
        return doc


def test_curried_nlp_pipe():
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences(name="my-sentences"))
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.sections(), name="sections")
    pipe = CustomComponent()

    err = (
        f"This component ({pipe}) has not been instantiated yet, likely because it was"
        " missing an `nlp` pipeline argument. You should either:\n"
        "- add it to a pipeline: `pipe = nlp.add_pipe(pipe)`\n"
        "- or fill its `nlp` argument: `pipe = factory(nlp=nlp, ...)`"
    )
    assert isinstance(pipe, CurriedFactory)
    with pytest.raises(TypeError) as exc_info:
        pipe("Demo texte")
    assert str(exc_info.value) == err

    with pytest.raises(TypeError) as exc_info:
        pipe.forward("Demo texte")
    assert str(exc_info.value) == err

    nlp.add_pipe(pipe, name="custom")

    assert nlp.pipes.custom.nlp is nlp

    assert nlp.pipe_names == ["my-sentences", "normalizer", "sections", "custom"]



from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Union,
)

from pydantic import NonNegativeInt
from spacy.tokens import Doc, Span
from typing_extensions import Literal

from edsnlp import registry
from edsnlp.utils.filter import filter_spans
from edsnlp.utils.typing import AsList

SeqStr = AsList[str]
SpanFilter = Union[bool, SeqStr]

SpanSetterMapping = Dict[str, SpanFilter]
SpanGetterMapping = Dict[str, SpanFilter]

SpanGetter = Union[
    SpanGetterMapping,
    Callable[[Doc], Iterable[Span]],
]
SpanSetter = Union[
    SpanSetterMapping,
    Callable[[Doc, Iterable[Span]], Any],
]


def get_spans(doc, span_getter):
    if span_getter is None:
        yield doc[:]
        return
    if callable(span_getter):
        yield from span_getter(doc)
        return
    for key, span_filter in span_getter.items():
        if key == "*":
            candidates = (span for group in doc.spans.values() for span in group)
        else:
            candidates = doc.spans.get(key, ()) if key != "ents" else doc.ents
        if span_filter is True:
            yield from candidates
        else:
            for span in candidates:
                if span.label_ in span_filter:
                    yield span


def get_spans_with_group(doc, span_getter):
    if span_getter is None:
        yield doc[:], None
        return
    if callable(span_getter):
        yield from span_getter(doc)
        return
    for key, span_filter in span_getter.items():
        if key == "*":
            candidates = (
                (span, group) for group in doc.spans.values() for span in group
            )
        else:
            candidates = doc.spans.get(key, ()) if key != "ents" else doc.ents
            candidates = ((span, key) for span in candidates)
        if span_filter is True:
            yield from candidates
        else:
            for span, group in candidates:
                if span.label_ in span_filter:
                    yield span, group


def set_spans(doc, matches, span_setter):
    if callable(span_setter):
        span_setter(doc, matches)
    else:
        match_all = []
        label_to_group = defaultdict(list)
        for name, spans_filter in span_setter.items():
            if name != "ents" and name != "*":
                doc.spans.setdefault(name, [])
            if spans_filter:
                if spans_filter is True:
                    match_all.append(name)
                else:
                    for label in spans_filter:
                        label_to_group[label].append(name)

        new_ents = [] if "ents" in span_setter else None

        for span in matches:
            for group in match_all + label_to_group[span.label_]:
                if group == "ents":
                    new_ents.append(span)
                elif group == "*":
                    doc.spans.setdefault(span.label_, []).append(span)
                else:
                    doc.spans[group].append(span)
        if new_ents is not None:
            doc.ents = filter_spans((*new_ents, *doc.ents))
    return doc


def validate_span_setter(value: Union[SeqStr, Dict[str, SpanFilter]]) -> SpanSetter:
    if callable(value):
        return value
    if isinstance(value, str):
        return {value: True}
    if isinstance(value, list):
        return {group: True for group in value}
    elif isinstance(value, dict):
        new_value = {}
        for k, v in value.items():
            if isinstance(v, bool):
                new_value[k] = v
            elif isinstance(v, str):
                new_value[k] = [v]
            elif isinstance(v, list) and all(isinstance(i, str) for i in v):
                new_value[k] = v
            else:
                raise TypeError(
                    f"Invalid entry {value} ({type(value)}) for SpanSetterArg, "
                    f"expected bool/string(s), dict of bool/string(s) or callable"
                )
        return new_value
    else:
        raise TypeError(
            f"Invalid entry {value} ({type(value)}) for SpanSetterArg, "
            f"expected bool/string(s), dict of bool/string(s) or callable"
        )


def validate_span_getter(
    value: Union[SeqStr, Dict[str, SpanFilter]], optional: bool = False
) -> SpanSetter:
    if value is None:
        if optional:
            return None
        raise ValueError(
            "Mising entry for SpanGetterArg, "
            "expected bool/string(s), dict of bool/string(s) or callable"
        )
    if callable(value):
        return value
    if isinstance(value, str):
        return {value: True}
    if isinstance(value, list):
        return {group: True for group in value}
    elif isinstance(value, dict):
        new_value = {}
        for k, v in value.items():
            if isinstance(v, bool):
                new_value[k] = v
            elif isinstance(v, str):
                new_value[k] = [v]
            elif isinstance(v, list) and all(isinstance(i, str) for i in v):
                new_value[k] = v
            else:
                raise TypeError(
                    f"Invalid entry {value} ({type(value)}) for SpanGetterArg, "
                    f"expected bool/string(s), dict of bool/string(s) or callable"
                )
        return new_value
    else:
        raise TypeError(
            f"Invalid entry {value} ({type(value)}) for SpanGetterArg, "
            f"expected bool/string(s), dict of bool/string(s) or callable"
        )


class SpanSetterArg:
    """
    Valid values for the `span_setter` argument of a component can be :

    - a (doc, matches) -> None callable
    - a span group name
    - a list of span group names
    - a dict of group name to True or list of labels

    The group name `"ents"` is a special case, and will add the matches to `doc.ents`

    Examples
    --------
    - `span_setter=["ents", "ckd"]` will add the matches to both `doc.ents` and
    `doc.spans["ckd"]`. It is equivalent to `{"ents": True, "ckd": True}`.
    - `span_setter={"ents": ["foo", "bar"]}` will add the matches with label
    "foo" and "bar" to `doc.ents`.
    - `span_setter="ents"` will add all matches only to `doc.ents`.
    - `span_setter="ckd"` will add all matches only to `doc.spans["ckd"]`.
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value, config=None) -> SpanSetter:
        return validate_span_setter(value)


class SpanGetterArg:
    """
    Valid values for the `span_getter` argument of a component can be :

    - a (doc) -> spans callable
    - a span group name
    - a list of span group names
    - a dict of group name to True or list of labels

    The group name `"ents"` is a special case, and will get the matches from `doc.ents`

    Examples
    --------
    - `span_getter=["ents", "ckd"]` will get the matches from both `doc.ents` and
    `doc.spans["ckd"]`. It is equivalent to `{"ents": True, "ckd": True}`.
    - `span_getter={"ents": ["foo", "bar"]}` will get the matches with label
    "foo" and "bar" from `doc.ents`.
    - `span_getter="ents"` will get all matches from `doc.ents`.
    - `span_getter="ckd"` will get all matches from `doc.spans["ckd"]`.
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value, config=None) -> SpanSetter:
        return validate_span_setter(value)


if TYPE_CHECKING:
    SpanGetterArg = Union[  # noqa: F811
        str,
        Sequence[str],
        SpanGetterMapping,
        Callable[[Doc], Iterable[Span]],
    ]
    SpanSetterArg = Union[  # noqa: F811
        str,
        Sequence[str],
        SpanSetterMapping,
        Callable[[Doc, Iterable[Span]], Any],
    ]


def merge_spans(spans: Iterable[Span]) -> List[Span]:
    """
    Merge overlapping spans into a single span.

    Parameters
    ----------
    spans : List[Span]
        List of spans to merge.
    doc : Doc
        Document to merge the spans on.

    Returns
    -------
    List[Span]
        Merged spans.
    """
    spans = sorted(spans, key=lambda x: (x.start, x.end))
    merged = []
    for span in spans:
        if len(merged) and span.start <= merged[-1].end:
            if span.end > merged[-1].end:
                merged[-1] = Span(
                    span.doc,
                    merged[-1].start,
                    span.end,
                    merged[-1].label_,
                )
        else:
            merged.append(span)
    return merged


@registry.misc.register("eds.span_context_getter")
class make_span_context_getter:
    """
    Create a span context getter.

    Parameters
    ----------
    span_getter : SpanGetterArg
        Span getter, i.e. for which spans to get the context.
    context_words : NonNegativeInt
        Minimum number of words to include on each side of the span.
    context_sents : Optional[NonNegativeInt]
        Minimum number of sentences to include on each side of the span:

        - 0: don't use sentences to build the context.
        - 1: include the sentence of the span.
        - n: include n sentences on each side of the span.

        By default, 0 if the document has no sentence annotations, 1 otherwise.
    overlap_policy : Literal["filter", "merge"]
        How to handle overlapping spans:

        - "filter": remove overlapping spans.
        - "merge": merge overlapping spans
    """

    def __init__(
        self,
        span_getter: SpanGetterArg,
        context_words: NonNegativeInt = 0,
        context_sents: Optional[NonNegativeInt] = None,
        overlap_policy: Literal["filter", "merge"] = "merge",
    ):
        self.context_words = context_words
        self.context_sents = context_sents
        self.overlap_policy = overlap_policy
        self.span_getter = span_getter

    def __call__(self, doc: Doc) -> List[Span]:
        n_sents = self.context_sents
        if n_sents is None:
            n_sents = 0 if not doc.has_annotation("SENT_START") else 1
        n_words = self.context_words

        spans = []
        sents = list(doc.sents) if n_sents > 1 else []
        for e in get_spans(doc, self.span_getter):
            min_start_sent = min_start_word = e.start - n_words
            max_end_sent = max_end_word = e.end + n_words

            if n_sents == 1:
                sent = e.sent
                min_start_sent = sent.start
                max_end_sent = sent.end
            else:
                sent_i = sents.index(e.sent)
                min_start_sent = sents[max(0, sent_i - n_sents)].start
                max_end_sent = sents[min(len(sents) - 1, sent_i + n_sents)].end
            start = max(0, min(min_start_word, min_start_sent))
            end = min(len(doc), max(max_end_word, max_end_sent))
            spans.append(doc[start:end])

        if self.overlap_policy == "filter":
            return filter_spans(spans)
        return merge_spans(spans)



"""`eds.ckd` pipeline"""
from typing import Any, Dict, List, Optional, Union

from loguru import logger
from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.pipes.ner.disorders.base import DisorderMatcher

from .patterns import default_patterns


class CKDMatcher(DisorderMatcher):
    """

    The `eds.CKD` pipeline component extracts mentions of CKD (Chronic Kidney Disease).
    It will notably match:

    - Mentions of various diseases (see below)
    - Kidney transplantation
    - Chronic dialysis
    - Renal failure **from stage 3 to 5**. The stage is extracted by trying 3 methods:
        - Extracting the mentioned stage directly ("*IRC stade IV*")
        - Extracting the severity directly ("*IRC terminale*")
        - Extracting the mentioned GFR (DFG in french) ("*IRC avec DFG estimé à 30
          mL/min/1,73m2)*")

    ??? info "Details of the used patterns"
        ```{ .python .no-check }
        # fmt: off
        --8<-- "edsnlp/pipes/ner/disorders/CKD/patterns.py"
        # fmt: on
        ```

    Extensions
    ----------
    On each span `span` that match, the following attributes are available:

    - `span._.detailed_status`: set to `"PRESENT"`
    - `span._.assigned`: dictionary with the following keys, if relevant:
        - `stage`: mentioned renal failure stage
        - `status`: mentioned renal failure severity (e.g. modérée, sévère, terminale,
          etc.)
        - `dfg`: mentioned DFG

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.normalizer(
            accents=True,
            lowercase=True,
            quotes=True,
            spaces=True,
            pollution=dict(
                information=True,
                bars=True,
                biology=True,
                doctors=True,
                web=True,
                coding=True,
                footer=True,
            ),
        ),
    )
    nlp.add_pipe(eds.ckd())
    ```

    Below are a few examples:

    --8<-- "docs/assets/fragments/ckd-examples.md"

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline
    name : Optional[str]
        The name of the component
    patterns: Union[Dict[str, Any], List[Dict[str, Any]]]
        The patterns to use for matching
    label : str
        The label to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Authors and citation
    --------------------
    The `eds.CKD` component was developed by AP-HP's Data Science team with a team of
    medical experts. A paper describing in details the development of those components
    is being drafted and will soon be available.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: str = "ckd",
        *,
        patterns: Union[Dict[str, Any], List[Dict[str, Any]]] = default_patterns,
        label: str = "ckd",
        span_setter: SpanSetterArg = {"ents": True, "ckd": True},
    ):
        super().__init__(
            nlp=nlp,
            name=name,
            label=label,
            patterns=patterns,
            span_setter=span_setter,
        )

    def classify_from_dfg(self, dfg_span: Optional[Span]):
        if dfg_span is None:
            return False
        try:
            dfg_value = float(dfg_span.text.replace(",", ".").strip())
        except ValueError:
            logger.trace(f"DFG value couldn't be extracted from {dfg_span.text}")
            return False

        return dfg_value < 60  # We keep only moderate to severe CKD

    def process(self, doc: Doc):
        for span in super().process(doc):
            if span._.source == "dialysis" and "chronic" not in span._.assigned.keys():
                continue

            if span._.source == "general":
                if {"stage", "status"} & set(span._.assigned.keys()):
                    yield span
                elif self.classify_from_dfg(span._.assigned.get("dfg", None)):
                    yield span
            else:
                yield span



from ..terms import HEART

main_pattern = dict(
    source="main",
    regex=[
        r"coronaropathie",
        r"angor.{1,5}instable",
        r"cardiopathie(?!.{0,20}non).{0,20}(ischem|arteriosc)",
        r"cardio.?myopathie(?!.{0,20}non).{0,20}(ischem|arteriosc)",
        r"ischemi.{1,15}myocard",
        r"syndrome.{1,5}corona.{1,10}aigu",
        r"syndrome.{1,5}corona.{1,10}st",
        r"pontage.{1,5}mammaire",
    ],
    regex_attr="NORM",
)

with_localization = dict(
    source="with_localization",
    regex=[
        r"\bstent",
        r"endoprothese",
        r"pontage",
        r"anevr[iy]sme",
        "infarctus",
        r"angioplasti",
    ],
    assign=[
        dict(
            name="heart_localized",
            regex="(" + r"|".join(HEART) + ")",
            window=(-10, 10),
        ),
    ],
    regex_attr="NORM",
)

acronym = dict(
    source="acronym",
    regex=[
        r"\bidm\b",
        r"\bsca\b",
        r"\batl\b",
    ],
    regex_attr="NORM",
    assign=dict(
        name="segment",
        regex=r"st([+-])",
        window=2,
    ),
)


default_patterns = [
    main_pattern,
    with_localization,
    acronym,
]



# ruff: noqa: E501
"""
# -----------
VENDORED https://github.com/mkdocstrings/autorefs/blob/e19b9fa47dac136a529c2be0d7969106ca5d5106/src/mkdocs_autorefs/
Waiting for the following PR to be merged: https://github.com/mkdocstrings/autorefs/pull/25
# -----------

This module contains the "mkdocs-autorefs" plugin.

After each page is processed by the Markdown converter, this plugin stores absolute URLs of every HTML anchors
it finds to later be able to fix unresolved references.
It stores them during the [`on_page_content` event hook](https://www.mkdocs.org/user-guide/plugins/#on_page_content).

Just before writing the final HTML to the disc, during the
[`on_post_page` event hook](https://www.mkdocs.org/user-guide/plugins/#on_post_page),
this plugin searches for references of the form `[identifier][]` or `[title][identifier]` that were not resolved,
and fixes them using the previously stored identifier-URL mapping.
"""
import contextlib
import functools
import logging
import os
import re
from html import escape, unescape
from typing import Any, Callable, Dict, List, Match, Optional, Sequence, Tuple, Union
from urllib.parse import urlsplit
from xml.etree.ElementTree import Element

from markdown import Markdown
from markdown.extensions import Extension
from markdown.inlinepatterns import REFERENCE_RE, ReferenceInlineProcessor
from markdown.util import INLINE_PLACEHOLDER_RE
from mkdocs.config import Config
from mkdocs.config import config_options as c
from mkdocs.plugins import BasePlugin
from mkdocs.structure.pages import Page
from mkdocs.structure.toc import AnchorLink
from mkdocs.utils import warning_filter

AUTO_REF_RE = re.compile(
    r"<span data-(?P<kind>autorefs-identifier|autorefs-optional|autorefs-optional-hover)="
    r'("?)(?P<identifier>[^"<>]*)\2>(?P<title>.*?)</span>'
)
"""A regular expression to match mkdocs-autorefs' special reference markers
in the [`on_post_page` hook][mkdocs_autorefs.plugin.AutorefsPlugin.on_post_page].
"""

EvalIDType = Tuple[Any, Any, Any]


class AutoRefInlineProcessor(ReferenceInlineProcessor):
    """A Markdown extension."""

    def __init__(self, *args, **kwargs):  # noqa: D107
        super().__init__(REFERENCE_RE, *args, **kwargs)

    # Code based on
    # https://github.com/Python-Markdown/markdown/blob/8e7528fa5c98bf4652deb13206d6e6241d61630b/markdown/inlinepatterns.py#L780

    def handleMatch(self, m, data) -> Union[Element, EvalIDType]:  # type: ignore[override]  # noqa: N802,WPS111
        """Handle an element that matched.

        Arguments:
            m: The match object.
            data: The matched data.

        Returns:
            A new element or a tuple.
        """
        text, index, handled = self.getText(data, m.end(0))
        if not handled:
            return None, None, None

        identifier, end, handled = self.evalId(data, index, text)
        if not handled:
            return None, None, None

        if re.search(r"[/ \x00-\x1f]", identifier):
            # Do nothing if the matched reference contains:
            # - a space, slash or control character (considered unintended);
            # - specifically \x01 is used by Python-Markdown HTML stash when there's inline formatting,
            #   but references with Markdown formatting are not possible anyway.
            return None, m.start(0), end

        return self.makeTag(identifier, text), m.start(0), end

    def evalId(
        self, data: str, index: int, text: str
    ) -> EvalIDType:  # noqa: N802 (parent's casing)
        """Evaluate the id portion of `[ref][id]`.

        If `[ref][]` use `[ref]`.

        Arguments:
            data: The data to evaluate.
            index: The starting position.
            text: The text to use when no identifier.

        Returns:
            A tuple containing the identifier, its end position, and whether it matched.
        """
        m = self.RE_LINK.match(data, pos=index)  # noqa: WPS111
        if not m:
            return None, index, False

        identifier = m.group(1)
        if not identifier:
            identifier = text
            # Allow the entire content to be one placeholder, with the intent of catching things like [`Foo`][].
            # It doesn't catch [*Foo*][] though, just due to the priority order.
            # https://github.com/Python-Markdown/markdown/blob/1858c1b601ead62ed49646ae0d99298f41b1a271/markdown/inlinepatterns.py#L78
            if INLINE_PLACEHOLDER_RE.fullmatch(identifier):
                identifier = self.unescape(identifier)

        end = m.end(0)
        return identifier, end, True

    def makeTag(self, identifier: str, text: str) -> Element:  # type: ignore[override]  # noqa: N802,W0221
        """Create a tag that can be matched by `AUTO_REF_RE`.

        Arguments:
            identifier: The identifier to use in the HTML property.
            text: The text to use in the HTML tag.

        Returns:
            A new element.
        """
        el = Element("span")
        el.set("data-autorefs-identifier", identifier)
        el.text = text
        return el


def relative_url(url_a: str, url_b: str) -> str:
    """Compute the relative path from URL A to URL B.

    Arguments:
        url_a: URL A.
        url_b: URL B.

    Returns:
        The relative URL to go from A to B.
    """
    parts_a = url_a.split("/")
    url_b, anchor = url_b.split("#", 1)
    parts_b = url_b.split("/")

    # remove common left parts
    while parts_a and parts_b and parts_a[0] == parts_b[0]:
        parts_a.pop(0)
        parts_b.pop(0)

    # go up as many times as remaining a parts' depth
    levels = len(parts_a) - 1
    parts_relative = [".."] * levels + parts_b  # noqa: WPS435
    relative = "/".join(parts_relative)
    return f"{relative}#{anchor}"


def fix_ref(
    url_mapper: Callable[[str], str], unmapped: List[str]
) -> Callable:  # noqa: WPS212,WPS231
    """Return a `repl` function for [`re.sub`](https://docs.python.org/3/library/re.html#re.sub).

    In our context, we match Markdown references and replace them with HTML links.

    When the matched reference's identifier was not mapped to an URL, we append the identifier to the outer
    `unmapped` list. It generally means the user is trying to cross-reference an object that was not collected
    and rendered, making it impossible to link to it. We catch this exception in the caller to issue a warning.

    Arguments:
        url_mapper: A callable that gets an object's site URL by its identifier,
            such as [mkdocs_autorefs.plugin.AutorefsPlugin.get_item_url][].
        unmapped: A list to store unmapped identifiers.

    Returns:
        The actual function accepting a [`Match` object](https://docs.python.org/3/library/re.html#match-objects)
        and returning the replacement strings.
    """

    def inner(match: Match):  # noqa: WPS212,WPS430
        identifier = match["identifier"]
        title = match["title"]
        kind = match["kind"]

        try:
            url = url_mapper(unescape(identifier))
        except KeyError:
            if kind == "autorefs-optional":
                return title
            elif kind == "autorefs-optional-hover":
                return f'<span title="{identifier}">{title}</span>'
            unmapped.append(identifier)
            if title == identifier:
                return f"[{identifier}][]"
            return f"[{title}][{identifier}]"

        parsed = urlsplit(url)
        external = parsed.scheme or parsed.netloc
        classes = ["autorefs", "autorefs-external" if external else "autorefs-internal"]
        class_attr = " ".join(classes)
        if kind == "autorefs-optional-hover":
            return f'<a class="{class_attr}" title="{identifier}" href="{escape(url)}">{title}</a>'
        return f'<a class="{class_attr}" href="{escape(url)}">{title}</a>'

    return inner


def fix_refs(html: str, url_mapper: Callable[[str], str]) -> Tuple[str, List[str]]:
    """Fix all references in the given HTML text.

    Arguments:
        html: The text to fix.
        url_mapper: A callable that gets an object's site URL by its identifier,
            such as [mkdocs_autorefs.plugin.AutorefsPlugin.get_item_url][].

    Returns:
        The fixed HTML.
    """
    unmapped = []  # type: ignore
    html = AUTO_REF_RE.sub(fix_ref(url_mapper, unmapped), html)
    return html, unmapped


class AutorefsExtension(Extension):
    """Extension that inserts auto-references in Markdown."""

    def extendMarkdown(
        self, md: Markdown
    ) -> None:  # noqa: N802 (casing: parent method's name)
        """Register the extension.

        Add an instance of our [`AutoRefInlineProcessor`][mkdocs_autorefs.references.AutoRefInlineProcessor] to the Markdown parser.

        Arguments:
            md: A `markdown.Markdown` instance.
        """
        md.inlinePatterns.register(
            AutoRefInlineProcessor(md),
            "mkdocs-autorefs",
            priority=168,  # noqa: WPS432  # Right after markdown.inlinepatterns.ReferenceInlineProcessor
        )


log = logging.getLogger(f"mkdocs.plugins.{__name__}")
log.addFilter(warning_filter)


class AutorefsPlugin(BasePlugin):
    """An `mkdocs` plugin.

    This plugin defines the following event hooks:

    - `on_config`
    - `on_page_content`
    - `on_post_page`

    Check the [Developing Plugins](https://www.mkdocs.org/user-guide/plugins/#developing-plugins) page of `mkdocs`
    for more information about its plugin system.
    """

    scan_toc: bool = True
    current_page: Optional[str] = None
    config_scheme = (("priority", c.ListOfItems(c.Type(str), default=[])),)

    def __init__(self) -> None:
        """Initialize the object."""
        super().__init__()
        self._url_map: Dict[str, str] = {}
        self._abs_url_map: Dict[str, str] = {}
        self.get_fallback_anchor: Optional[
            Callable[[str], Optional[str]]
        ] = None  # noqa: WPS234
        self._priority_patterns = None

    @property
    def priority_patterns(self):
        if self._priority_patterns is None:
            self._priority_patterns = [
                os.path.join("/", pat) for pat in self.config.get("priority")
            ]
        return self._priority_patterns

    def register_anchor(self, url: str, identifier: str):
        """Register that an anchor corresponding to an identifier was encountered when rendering the page.

        Arguments:
            url: The relative URL of the current page. Examples: `'foo/bar/'`, `'foo/index.html'`
            identifier: The HTML anchor (without '#') as a string.
        """

        new_url = os.path.join("/", f"{url}#{identifier}")
        old_url = os.path.join("/", self._url_map.get(identifier, "")).split("#")[0]

        if identifier in self._url_map and not old_url == new_url:
            rev_patterns = list(enumerate(self.priority_patterns))[::-1]
            old_priority_idx = next(
                (i for i, pat in rev_patterns if re.match(pat, old_url)),
                len(rev_patterns),
            )
            new_priority_idx = next(
                (i for i, pat in rev_patterns if re.match(pat, new_url)),
                len(rev_patterns),
            )
            if new_priority_idx >= old_priority_idx:
                return
            if "reference" not in new_url:
                raise Exception("URL WTF", new_url)

        self._url_map[identifier] = new_url

    def register_url(self, identifier: str, url: str):
        """Register that the identifier should be turned into a link to this URL.

        Arguments:
            identifier: The new identifier.
            url: The absolute URL (including anchor, if needed) where this item can be found.
        """
        self._abs_url_map[identifier] = url

    def _get_item_url(  # noqa: WPS234
        self,
        identifier: str,
        fallback: Optional[Callable[[str], Sequence[str]]] = None,
    ) -> str:
        try:
            return self._url_map[identifier]
        except KeyError:
            if identifier in self._abs_url_map:
                return self._abs_url_map[identifier]
            if fallback:
                new_identifiers = fallback(identifier)
                for new_identifier in new_identifiers:
                    with contextlib.suppress(KeyError):
                        url = self._get_item_url(new_identifier)
                        self._url_map[identifier] = url
                        return url
            raise

    def get_item_url(  # noqa: WPS234
        self,
        identifier: str,
        from_url: Optional[str] = None,
        fallback: Optional[Callable[[str], Sequence[str]]] = None,
    ) -> str:
        """Return a site-relative URL with anchor to the identifier, if it's present anywhere.

        Arguments:
            identifier: The anchor (without '#').
            from_url: The URL of the base page, from which we link towards the targeted pages.
            fallback: An optional function to suggest alternative anchors to try on failure.

        Returns:
            A site-relative URL.
        """
        return self._get_item_url(identifier, fallback)

    def on_config(
        self, config: Config, **kwargs
    ) -> Config:  # noqa: W0613,R0201 (unused arguments, cannot be static)
        """Instantiate our Markdown extension.

        Hook for the [`on_config` event](https://www.mkdocs.org/user-guide/plugins/#on_config).
        In this hook, we instantiate our [`AutorefsExtension`][mkdocs_autorefs.references.AutorefsExtension]
        and add it to the list of Markdown extensions used by `mkdocs`.

        Arguments:
            config: The MkDocs config object.
            kwargs: Additional arguments passed by MkDocs.

        Returns:
            The modified config.
        """
        log.debug(f"{__name__}: Adding AutorefsExtension to the list")
        config["markdown_extensions"].append(AutorefsExtension())
        return config

    def on_page_markdown(
        self, markdown: str, page: Page, **kwargs
    ) -> str:  # noqa: W0613 (unused arguments)
        """Remember which page is the current one.

        Arguments:
            markdown: Input Markdown.
            page: The related MkDocs page instance.
            kwargs: Additional arguments passed by MkDocs.

        Returns:
            The same Markdown. We only use this hook to map anchors to URLs.
        """
        self.current_page = page.url  # noqa: WPS601
        return markdown

    def on_page_content(
        self, html: str, page: Page, **kwargs
    ) -> str:  # noqa: W0613 (unused arguments)
        """Map anchors to URLs.

        Hook for the [`on_page_content` event](https://www.mkdocs.org/user-guide/plugins/#on_page_content).
        In this hook, we map the IDs of every anchor found in the table of contents to the anchors absolute URLs.
        This mapping will be used later to fix unresolved reference of the form `[title][identifier]` or
        `[identifier][]`.

        Arguments:
            html: HTML converted from Markdown.
            page: The related MkDocs page instance.
            kwargs: Additional arguments passed by MkDocs.

        Returns:
            The same HTML. We only use this hook to map anchors to URLs.
        """
        if self.scan_toc:
            log.debug(
                f"{__name__}: Mapping identifiers to URLs for page {page.file.src_path}"
            )
            for item in page.toc.items:
                self.map_urls(page, item)
        return html

    def map_urls(self, page: Page, anchor: AnchorLink) -> None:
        """Recurse on every anchor to map its ID to its absolute URL.

        This method populates `self.url_map` by side-effect.

        Arguments:
            base_url: The base URL to use as a prefix for each anchor's relative URL.
            anchor: The anchor to process and to recurse on.
        """
        abs_url = os.path.join("/", page.file.url)
        self.register_anchor(abs_url, anchor.id)
        for child in anchor.children:
            self.map_urls(page, child)

    def on_post_page(
        self, output: str, page: Page, **kwargs
    ) -> str:  # noqa: W0613 (unused arguments)
        """Fix cross-references.

        Hook for the [`on_post_page` event](https://www.mkdocs.org/user-guide/plugins/#on_post_page).
        In this hook, we try to fix unresolved references of the form `[title][identifier]` or `[identifier][]`.
        Doing that allows the user of `autorefs` to cross-reference objects in their documentation strings.
        It uses the native Markdown syntax so it's easy to remember and use.

        We log a warning for each reference that we couldn't map to an URL, but try to be smart and ignore identifiers
        that do not look legitimate (sometimes documentation can contain strings matching
        our [`AUTO_REF_RE`][mkdocs_autorefs.references.AUTO_REF_RE] regular expression that did not intend to reference anything).
        We currently ignore references when their identifier contains a space or a slash.

        Arguments:
            output: HTML converted from Markdown.
            page: The related MkDocs page instance.
            kwargs: Additional arguments passed by MkDocs.

        Returns:
            Modified HTML.
        """
        log.debug(f"{__name__}: Fixing references in page {page.file.src_path}")

        url_mapper = functools.partial(
            self.get_item_url, from_url=page.url, fallback=self.get_fallback_anchor
        )
        fixed_output, unmapped = fix_refs(output, url_mapper)

        if unmapped and log.isEnabledFor(logging.WARNING):
            for ref in unmapped:
                log.warning(
                    f"{__name__}: {page.file.src_path}: Could not find cross-reference target '[{ref}]'",
                )

        return fixed_output



from typing import Optional

from spacy.tokens import Doc

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import BaseComponent


class SpacesTagger(BaseComponent):
    """
    We assign "SPACE" to `token.tag` to be used by optimized components
    such as the EDSPhraseMatcher

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline object.
    name : Optional[str]
        The component name.
    newline : bool
        Whether to update the newline tokens too
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol] = None,
        name: Optional[str] = "spaces",
        *,
        newline: bool = True,
    ):
        super().__init__(nlp, name)
        self.newline = newline

    def __call__(self, doc: Doc) -> Doc:
        """
        Apply the component to the doc.

        Parameters
        ----------
        doc: Doc

        Returns
        -------
        doc: Doc
        """
        space_hash = doc.vocab.strings["SPACE"]
        for token in doc:
            if len(token.text.strip()) == 0:
                token.tag = space_hash

        return doc



import polars

import edsnlp


def test_read_write(blank_nlp, text, df_notes_pandas):
    df_notes_polars = polars.from_pandas(df_notes_pandas)
    reader = edsnlp.data.from_polars(
        df_notes_polars,
        converter="omop",
        nlp=blank_nlp,
    ).set_processing(backend="simple")
    doc = list(reader)[0]
    assert doc.text == text

    blank_nlp.add_pipe("eds.matcher", config={"terms": {"douleur": ["douleurs"]}})
    blank_nlp.add_pipe("eds.negation")
    docs = reader.map_pipeline(blank_nlp)

    writer: polars.DataFrame = docs.to_polars(
        converter="omop",
        span_attributes=["negation"],
        span_getter=["ents"],
    )
    res = writer.to_dicts()
    assert len(res) == 20
    assert sum(len(r["entities"]) for r in res) == 20



PA = r"(?:\bp/?a\b|paquets?.?annee)"
QUANTITY = r"(?P<quantity>[\d]{1,3})"
PUNCT = r"\.,-;\(\)"

default_patterns = [
    dict(
        source="tobacco",
        regex=[
            r"tabagi",
            r"tabac",
            r"\bfume\b",
            r"\bfumeu",
            r"\bpipes?\b",
        ],
        exclude=dict(
            regex=[
                "occasion",
                "moder",
                "quelqu",
                "festi",
                "rare",
                "sujet",  # Example : Chez le sujet fumeur ... generic sentences
            ],
            window=(-3, 5),
        ),
        regex_attr="NORM",
        assign=[
            dict(
                name="stopped",
                regex=r"(?<!non )(?<!pas )(\bex\b|sevr|arret|stop|ancien)",
                window=(-3, 15),
            ),
            dict(
                name="zero_after",
                regex=r"(?=^[a-z]*\s*:?[\s-]*(0|non(?! sevr)))",
                window=6,
            ),
            dict(
                name="PA",
                regex=rf"{QUANTITY}[^{PUNCT}]{{0,10}}{PA}|{PA}[^{PUNCT}]{{0,10}}{QUANTITY}",
                window=(-10, 10),
                reduce_mode="keep_first",
            ),
            dict(
                name="secondhand",
                regex="(passif)",
                window=5,
                reduce_mode="keep_first",
            ),
        ],
    )
]



import math
from enum import Enum
from typing import TypeVar

import torch

Args = TypeVar("Args")


def pad_2d(data, pad=0, **kwargs):
    max_len = max(map(len, data), default=0)
    padded = [row + [pad] * (max_len - len(row)) for row in data]
    return torch.as_tensor(padded, **kwargs)


class ActivationFunction(str, Enum):
    relu = "relu"
    gelu = "gelu"
    glu = "glu"


def get_activation_function(activation: ActivationFunction):
    return getattr(torch.nn.functional, activation)


def mask_to_triangle(mask):
    """
    Convert a boolean mask to a tensor containing distance to the nearest
    edge of the mask, in each direction.
    For example, if the mask is
    ```
    [
        [1, 1, 1, 1, 1],
        [1, 1, 1, 0, 0],
    ]
    ```

    The output will be
    ```
    [
        [0, 1, 2, 1, 0],
        [0, 1, 0, -1, -2]
    ]
    ```

    Parameters
    ----------
    mask: torch.Tensor
        A boolean mask

    Returns
    -------
    torch.Tensor
    """
    ramp = torch.arange(0, mask.shape[1], 1)
    scores = torch.min(ramp, mask.sum(1, keepdim=True) - 1 - ramp).view(-1)
    return scores


def make_windows(lengths, size, stride):
    max_len = max(lengths)
    windows = pad_2d(
        [
            list(
                range(
                    idx * stride + doc_idx * max_len,
                    min(idx * stride + size, length) + doc_idx * max_len,
                )
            )
            for doc_idx, length in enumerate(lengths)
            for idx in range(0, 1 + max(0, math.ceil((length - size) / stride)))
        ],
        pad=-1,
    )
    windows_mask = windows != -1
    windows[~windows_mask] = 0
    indexer = torch.zeros((len(lengths), max_len), dtype=torch.long).view(-1)
    scores = mask_to_triangle(windows_mask)
    scores = scores * len(scores) + torch.arange(len(scores))
    scores[~windows_mask.view(-1)] = -1
    indexer.index_reduce_(
        dim=0,
        source=scores,
        index=windows.view(-1),
        reduce="amax",
    )
    indexer %= len(scores)
    return windows.masked_fill(~windows_mask, -1), indexer



# Based on https://github.com/darwindarak/mdx_bib
import re
import string
from collections import Counter, OrderedDict
from typing import Tuple
from xml.etree import ElementTree as etree
from xml.etree.ElementTree import tostring as etree_to_string

from markdown.extensions import Extension
from markdown.inlinepatterns import Pattern
from markdown.preprocessors import Preprocessor
from mkdocs.config.config_options import Type as MkType
from mkdocs.plugins import BasePlugin
from pybtex.database.input import bibtex
from pybtex.exceptions import PybtexError

BRACKET_RE = re.compile(r"\[([^\[]+)\]")
CITE_RE = re.compile(r"@([\w_:-]+)")
DEF_RE = re.compile(r"\A {0,3}\[@([\w_:-]+)\]:\s*(.*)")
INDENT_RE = re.compile(r"\A\t| {4}(.*)")

CITATION_RE = r"(\[@(?:[\w_:-]+)(?: *, *@(?:[\w_:-]+))*\])"


class Bibliography(object):
    """Keep track of document references and citations for exporting"""

    def __init__(self, extension, plugin, bibtex_file, order):
        self.extension = extension
        self.order = order
        self.plugin = plugin

        self.citations = OrderedDict()
        self.references = dict()

        if bibtex_file:
            try:
                parser = bibtex.Parser()
                self.bibsource = parser.parse_file(bibtex_file).entries
                self.labels = {
                    id: self.formatCitation(self.bibsource[id])
                    for id in self.bibsource.keys()
                }
                for value, occurrences in Counter(self.labels.values()).items():
                    if occurrences > 1:
                        for xkey, xvalue in self.labels.items():
                            i = 0
                            if xvalue == value:
                                self.labels[
                                    xkey
                                ] = f"{xvalue}{string.ascii_lowercase[i]}"
                                i += 1

            except PybtexError:
                print("Error loading bibtex file")
                self.bibsource = dict()
                self.labels = {}
        else:
            self.bibsource = dict()

    def addCitation(self, citekey):
        self.citations[citekey] = self.citations.get(citekey, 0) + 1

    def setReference(self, citekey, reference):
        self.references[citekey] = reference

    def citationID(self, citekey):
        return "cite-" + citekey

    def referenceID(self, citekey):
        return "ref-" + citekey

    def formatAuthor(self, author):
        out = (
            author.last_names[0]
            + ((" " + author.first_names[0][0]) if author.first_names else "")
            + "."
        )
        if author.middle_names:
            out += f"{author.middle_names[0][0]}."
        return out.replace("{", "").replace("}", "")

    def formatAuthorSurname(self, author):
        out = author.last_names[0]
        return out.replace("{", "").replace("}", "")

    def formatReference(self, ref):
        author_list = list(map(self.formatAuthor, ref.persons["author"]))

        if len(author_list) == 1:
            authors = author_list[0]
        else:
            authors = ", ".join(author_list[:-1])
            authors += f" and {author_list[-1]}"

        # Harvard style
        # Surname, Initial, ... and Last_Surname,
        # Initial, Year. Title. Journal, Volume(Issue), pages. doi.

        title = ref.fields["title"].replace("{", "").replace("}", "")
        journal = ref.fields.get("journal", "")
        volume = ref.fields.get("volume", "")
        issue = ref.fields.get("issue", "")
        year = ref.fields.get("year")
        pages = ref.fields.get("pages")
        doi = ref.fields.get("doi")
        url = ref.fields.get("url")

        ref_id = self.referenceID(ref.key)
        ref = f"<p id={repr(ref_id)}>{authors}, {year}. {title}."
        if journal:
            ref += f" <i>{journal}</i>."
            if volume:
                ref += f" <i>{volume}</i>"
            if issue:
                ref += f"({issue})"
            if pages:
                ref += f", pp.{pages}"
            ref += "."
        if doi:
            ref += f' <a href="https://dx.doi.org/{doi}" target="_blank">{doi}</a>'
        elif url:
            ref += f' <a href="{url}" target="_blank">{url}</a>'
        ref += "</p>"

        return etree.fromstring(ref)

    def formatCitation(self, ref):
        author_list = list(map(self.formatAuthorSurname, ref.persons["author"]))
        year = ref.fields.get("year")

        if len(author_list) == 1:
            citation = f"{author_list[0]}"
        elif len(author_list) == 2:
            citation = f"{author_list[0]} and {author_list[1]}"
        else:
            citation = f"{author_list[0]} et al."

        citation += f", {year}"

        return citation

    def make_bibliography(self):
        if self.order == "alphabetical":
            raise (NotImplementedError)

        div = etree.Element("div")
        div.set("class", "footnote")
        div.append(etree.Element("hr"))
        ol = etree.SubElement(div, "ol")

        if not self.citations:
            return div

        # table = etree.SubElement(div, "table")
        # table.set("class", "references")
        # tbody = etree.SubElement(table, "tbody")
        etree.SubElement(div, "div")
        for id in self.citations:
            li = etree.SubElement(ol, "li")
            li.set("id", self.referenceID(id))
            # ref_id = etree.SubElement(li, "td")
            ref_txt = etree.SubElement(li, "p")
            if id in self.references:
                self.extension.parser.parseChunk(ref_txt, self.references[id])
            elif id in self.bibsource:
                ref_txt.append(self.formatReference(self.bibsource[id]))
            else:
                ref_txt.text = "Missing citation for {}".format(id)

        return div

    def clear_citations(self):
        self.citations = OrderedDict()


class CitationsPreprocessor(Preprocessor):
    """Gather reference definitions and citation keys"""

    def __init__(self, bibliography):
        self.bib = bibliography

    def subsequentIndents(self, lines, i):
        """Concatenate consecutive indented lines"""
        linesOut = []
        while i < len(lines):
            m = INDENT_RE.match(lines[i])
            if m:
                linesOut.append(m.group(1))
                i += 1
            else:
                break
        return " ".join(linesOut), i

    def run(self, lines):
        linesOut = []
        i = 0

        while i < len(lines):
            # Check to see if the line starts a reference definition
            m = DEF_RE.match(lines[i])
            if m:
                key = m.group(1)
                reference = m.group(2)
                indents, i = self.subsequentIndents(lines, i + 1)
                reference += " " + indents

                self.bib.setReference(key, reference)
                continue

            # Look for all @citekey patterns inside hard brackets
            for bracket in BRACKET_RE.findall(lines[i]):
                for c in CITE_RE.findall(bracket):
                    self.bib.addCitation(c)
            linesOut.append(lines[i])
            i += 1

        return linesOut


class CitationsPattern(Pattern):
    """Handles converting citations keys into links"""

    def __init__(self, pattern, bibliography):
        super(CitationsPattern, self).__init__(pattern)
        self.bib = bibliography

    def handleMatch(self, m):
        span = etree.Element("span")
        for cite_match in CITE_RE.finditer(m.group(2)):
            id = cite_match.group(1)
            if id in self.bib.bibsource:
                a = etree.Element("a")
                a.set("id", self.bib.citationID(id))
                a.set("href", "./#" + self.bib.referenceID(id))
                a.set("class", "citation")
                a.text = self.bib.labels[id]
                span.append(a)
            else:
                continue
        if len(span) == 0:
            return None
        return span


context_citations = None


class CitationsExtension(Extension):
    def __init__(self):
        super(CitationsExtension, self).__init__()
        self.bib = None

    def extendMarkdown(self, md):
        md.registerExtension(self)
        self.parser = md.parser
        self.md = md

        md.preprocessors.register(CitationsPreprocessor(self.bib), "mdx_bib", 15)
        md.inlinePatterns.register(
            CitationsPattern(CITATION_RE, self.bib), "mdx_bib", 175
        )


def makeExtension(*args, **kwargs):
    return CitationsExtension(*args, **kwargs)


class BibTexPlugin(BasePlugin):
    config_scheme: Tuple[Tuple[str, MkType]] = (
        ("bibtex_file", MkType(str)),  # type: ignore[assignment]
        ("order", MkType(str, default="unsorted")),  # type: ignore[assignment]
    )

    def __init__(self):
        self.citations = None

    def on_config(self, config, **kwargs):
        extension = CitationsExtension()
        self.bib = Bibliography(
            extension,
            self,
            self.config["bibtex_file"],
            self.config["order"],
        )
        extension.bib = self.bib
        config["markdown_extensions"].append(extension)

    def on_page_content(self, html, page, config, files):
        html += "\n" + etree_to_string(self.bib.make_bibliography()).decode()
        self.bib.clear_citations()
        return html
